# coding=utf-8
#
# Author: Ualter Otoni Pereira
# ualter.junior@gmail.com
#
import os
import logging
import sys
from os.path import expanduser
import signal
from pyawscp.Functions import Functions
from pyawscp.Utils import Style, Utils
from pyawscp.Config import Config
from pyawscp.PyEc2Cp import PyEc2CP
from pyawscp.PyElbCp import PyElbCP
from pyawscp.PyAwsShell import PyAwsShell

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)

signal.signal(signal.SIGINT, signal.SIG_DFL)

class PyAwsCp:
    def __init__(self):
        FILE_LOG = expanduser("~") + "/.pyawscp/" + "log"
        if not os.path.exists(FILE_LOG):
           os.makedirs(FILE_LOG)

        rootLogger = logging.getLogger()
        logFormatter = logging.Formatter("%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")

        fileHandler = logging.FileHandler("{0}/{1}.log".format(FILE_LOG, "py-awscp"))
        fileHandler.setFormatter(logFormatter)
        fileHandler.setLevel(logging.WARN)
        rootLogger.addHandler(fileHandler)

        #consoleHandler = logging.StreamHandler(sys.stdout)
        #consoleHandler.setFormatter(logFormatter)
        #consoleHandler.setLevel(logging.DEBUG)
        #rootLogger.addHandler(consoleHandler)

        self.start()

    def start(self):  
        if len(sys.argv) > 1:      
           if sys.argv[1] == "list" or sys.argv[1] == "help":
              Utils.clearScreen()
              Functions.showFunctions()
           elif sys.argv[1] == "-l" or sys.argv[1] == "-h":
              Utils.clearScreen()
              Functions.showFunctionsSummary()
           else:
              # One-way function execution
              s   = ""
              for arg in sys.argv:
                 s += arg
                 s += ","
              s = s[:len(s) - 1] 
              PyAwsShell().onewayFunction(s)  
           sys.exit()

        # Interactive 
        config = Config()
        config.interactive = True
        try:
           PyAwsShell().cmdloop()
        except KeyboardInterrupt:
           print("ERROR")
           pass   

def main():
    """Start Python AWS Shell Cockpit v0.5"""
    pyAwsCp = PyAwsCp()

if __name__ == '__main__':
    #pyAwsCp = PyAwsCp()
    main()





#!/usr/bin/env python3

import boto3
import logging
import sys
import json
from botocore.exceptions import ClientError
from datetime import datetime
from pyawscp.Functions import Functions
from pyawscp.Utils import Utils, Style
from pyawscp.PrettyTable import PrettyTable
from pyawscp.Emoticons import Emoticons
from pyawscp.Config import Config
from pyawscp.TableArgs import TableArgs
from pygments import highlight, lexers, formatters

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)

class PyEc2CP:
    config = None

    def __init__(self, config):
        self.config = config

    # A main function called from user
    # Check if a specific subnetId is Public looking up its Route Table for a IGW-* 0.0.0.0/0
    # arguments:
    #      subnetId: mandatory
    def subnetIsPublic(self):
        ec2      = self.botoSession().client('ec2')
        subnetId = self.config.commandArguments

        if not subnetId:
           resultTxt = "Where is the SubnetId? You didn't tell me which one " + Emoticons.ops() +  Style.RESET
           return Utils.formatResult(Functions.SUBNET_IS_PUBLIC,resultTxt, self.config, "", True)

        filters=[{'Name': 'association.subnet-id','Values': [subnetId]}, 
                 {'Name': 'route.destination-cidr-block','Values': ['0.0.0.0/0']},
                 {'Name': 'route.gateway-id','Values': ['igw-*']}]
        if self.config.awsTagsToFilter():
           for tag in self.config.awsTags: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]}
               ) 
        route_tables = ec2.describe_route_tables(Filters=filters)

        jsonResult = ""
        if self.config.printResults: 
           jsonResult = Utils.dictToJson(route_tables)

        routeTableIgw = route_tables['RouteTables']
        if len(routeTableIgw) >= 1:
           resultTxt = "Subnet " + Style.GREEN + subnetId + Style.RESET + " is " + Style.GREEN + "PUBLIC" +  Style.RESET + "\n"

           printSeparator = False
           for route in routeTableIgw[0]["Routes"]:
               if route["DestinationCidrBlock"] == '0.0.0.0/0':
                  if route["State"]:
                     if not printSeparator:
                        resultTxt = resultTxt + Utils.separator()
                        printSeparator = True
                     resultTxt = resultTxt + "\n"
                     resultTxt = resultTxt + ' '.ljust(2,' ') + "- State...........: " + Style.GREEN + route['State'] + Style.RESET 
                  if route["GatewayId"]:
                     if not printSeparator:
                        resultTxt = resultTxt + Utils.separator()
                        printSeparator = True
                     resultTxt = resultTxt + "\n"
                     resultTxt = resultTxt + ' '.ljust(2,' ') + "- GatewayId.......: " + Style.GREEN + route['GatewayId'] + Style.RESET    
                  if routeTableIgw[0]["VpcId"]:
                     if not printSeparator:
                        resultTxt = resultTxt + Utils.separator()
                        printSeparator = True
                     resultTxt = resultTxt + "\n"
                     resultTxt = resultTxt + ' '.ljust(2,' ') + "- Vpc Id..........: " + Style.GREEN + routeTableIgw[0]['VpcId'] + Style.RESET   
           return Utils.formatResult(Functions.FUNCTIONS[Functions.SUBNET_IS_PUBLIC]["name"],resultTxt, self.config, jsonResult, True, None)
        else:
           resultTxt = ""
           subnet = ec2.describe_subnets(Filters=[{'Name':'subnet-id','Values':[subnetId]}])
           if len(subnet['Subnets']) == 0:
               resultTxt = "Subnet " + Style.GREEN + subnetId + Style.RED + " not found! " + Emoticons.ops() +  Style.RESET
           else:   
               if self.config.awsTagsToFilter():
                  resultTxt = "Subnet " + Style.GREEN + subnetId + Style.RESET + " is NOT " + Style.GREEN + "PUBLIC" +  Style.RESET + ", or either did not pass the Filters Tag! " + + Emoticons.thumbsUp()
               else:   
                  resultTxt = "Subnet " + Style.GREEN + subnetId + Style.RESET + " is NOT " + Style.GREEN + "PUBLIC" +  Style.RESET + " " + Emoticons.thumbsUp() 
           return Utils.formatResult(Functions.FUNCTIONS[Functions.SUBNET_IS_PUBLIC]["name"],resultTxt, self.config, "", True, None)     
    
    # Reusable Function, used by: 
    # -lisSubtnetsVpc
    def _checkSubnetIsPublic(self, subnetId, extraFilters):
        ec2 = self.botoSession().client('ec2')
        filters=[{'Name': 'association.subnet-id','Values': [subnetId]}, 
                 {'Name': 'route.destination-cidr-block','Values': ['0.0.0.0/0']},
                 {'Name': 'route.gateway-id','Values': ['igw-*']}
        ]
        route_tables = ec2.describe_route_tables(Filters=filters)
        routeTableIgw = route_tables['RouteTables']
        if len(routeTableIgw) >= 1:
            return True
        else:
            return False    

    # List all Subnets from a VPC
    def listSubnetsVpc(self):
        ec2   = self.botoSession().client('ec2')

        vpcId     = ""
        isPublic  = False
        tableArgs = TableArgs()
        tableArgs.setArguments(self.config.commandArguments) 

        if "," in self.config.commandArguments:
           vpcId = self.config.commandArguments.split(",")[0]
           tableArgs.setArguments(self.config.commandArguments)
           if "ispublic" in self.config.commandArguments:
               isPublic = True
        else:
           vpcId = tableArgs.cleanPipelineArguments()

        if not vpcId:
           resultTxt = "Where is the VpcId? You didn't tell me which VPC " + Emoticons.ops() +  Style.RESET
           return Utils.formatResult(Functions.LIST_SUBNETS_VPC,resultTxt, self.config, "", True, tableArgs)

        filters=[
            {
                'Name': 'vpc-id',
                'Values': [
                    vpcId
                ]
            }
        ]
        if self.config.awsTagsToFilter():
           # Environment Tags set
           for tag in self.config.awsTags: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]}
               )

        # Tags from command line arguments
        if len(tableArgs.tagsTemp) > 0:
           for tag in tableArgs.tagsTemp: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[tableArgs.tagsTemp[tag]]}
               )

        subnets = ec2.describe_subnets( Filters=filters )

        jsonResult = ""
        if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile: 
           jsonResult = Utils.dictToJson(subnets)

        resultTxt = ""
        if len(subnets["Subnets"]) >= 1:
            resultTxt = "The VPC " + Style.GREEN + vpcId + Style.RESET + " has " + Style.GREEN + str(len(subnets["Subnets"])) + Style.RESET + " Subnets\n"
            resultTxt = resultTxt + Utils.separator()  + "\n\n"

            idx_lin=0
            header      = ["#","Subnet Id", "Availability Zone", "CIDR Block", "Available IP Address Count"]
            if isPublic:
                header.append("Is Public?")
            if tableArgs.showTags:
                header.append("Tags")    
            prettyTable = PrettyTable(header)
            
            for subnet in subnets["Subnets"]:
                idx_lin += 1

                columns = [ str(idx_lin), subnet['SubnetId'], subnet['AvailabilityZone'],  subnet['CidrBlock'], subnet['AvailableIpAddressCount'] ]

                if isPublic:
                   resultIsPublic = self._checkSubnetIsPublic(subnet['SubnetId'], None)
                   result = "--- "
                   if resultIsPublic:
                      result = "PUBLIC"
                   columns.append(result)
                if tableArgs.showTags:
                   tags = Utils.formatPrintTags(subnet['Tags'])
                   columns.append(tags)

                prettyTable.addRow(columns)

            if (int(tableArgs.sortCol) - 1) > len(columns):
                tableArgs.sortCol = "1"
            prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
            prettyTable.ascendingOrder(not tableArgs.desc)
            result = "VPC...: " + Style.GREEN + vpcId + Style.RESET + "\n\n" + prettyTable.printMe(self.config.tableLineSeparator, tableArgs)
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_SUBNETS_VPC]["name"], result, self.config, jsonResult, True, tableArgs)
        else:
           resultTxt = ""
           try: 
             vpc = ec2.describe_vpcs(VpcIds=[vpcId])
             if not self.config.awsTagsToFilter():
                resultTxt = "VPC was found, but there's no Subnet associated. " + Emoticons.ops()
             else:   
                resultTxt = "VPC was found, but there's no Subnet associated or either did not pass the Filters Tag! " + Emoticons.ops()
           except ClientError as e:
             if "InvalidVpcID.NotFound" == e.response['Error']['Code']:
                resultTxt = "VPC with the Id " + Style.GREEN + vpcId + Style.RESET + " was not found! " + Emoticons.ops() +  Style.RESET
             else:    
                resultTxt = "Exception! ❌💣 " + Style.GREEN + e.response['Error']['Code'] + Style.RESET
           except:
             resultTxt = sys.exc_info + " ❌💣 "
           return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_SUBNETS_VPC]["name"], resultTxt, self.config, "", True, tableArgs)

    # List all VPCs
    def listVpc(self):
        ec2 = self.botoSession().client('ec2')

        tableArgs = TableArgs()
        tableArgs.setArguments(self.config.commandArguments)

        listSubnets    = False
        subnetIsPublic = False

        if "subnets" in self.config.commandArguments:
           listSubnets    = True
        if "ispublic" in self.config.commandArguments:
           subnetIsPublic = True   

        filters = []
        if self.config.awsTagsToFilter():
           # Tags from Environment
           for tag in self.config.awsTags: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]}
               )

        # Tags from command line arguments
        if len(tableArgs.tagsTemp) > 0:
           for tag in tableArgs.tagsTemp: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[tableArgs.tagsTemp[tag]]}
               )
        vpcs = ec2.describe_vpcs(Filters=filters)

        jsonResult = ""
        if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile:
           jsonResult = Utils.dictToJson(vpcs)

        resultTxt = ""
        if len(vpcs["Vpcs"]) >= 1:
            resultTxt = "List of VPCs, total of " + Style.GREEN + str(len(vpcs["Vpcs"])) + Style.RESET + "\n"
            resultTxt = resultTxt + Utils.separator()  + "\n\n"

            idx_lin=0
            header = ["#","VPC Id", "CIDR Block", "Default"]
            if listSubnets:
               header = ["#","VPC Id", "CIDR Block", "Default", "Subnet Id", "Subnet CIDR Block", "AZ", "Available IP Count"] 
               if subnetIsPublic:
                  header.append("Is Public?")
            if tableArgs.showTags:
               header.append("VPC Tags")       

            prettyTable = PrettyTable(header)
            for vpc in vpcs["Vpcs"]:
                idx_lin += 1

                vpcId     = vpc['VpcId']
                cidrBlock = vpc['CidrBlock']
                default   = "Default" if vpc['IsDefault'] else "---"
                columns   = None
                # Asked to list the subnets of each VPC?
                if listSubnets:
                   subnets = ec2.describe_subnets( Filters=[{'Name': 'vpc-id','Values': [vpc['VpcId']]}])
                   idx_subnet_lin = 0

                   for subnet in subnets["Subnets"]:
                       idx_subnet_lin += 1
                       lblSubnet = subnet['SubnetId'] + "  (" + str(idx_subnet_lin) + ")"
                       cidrBlockSubnet = subnet['CidrBlock']
                       az = subnet['AvailabilityZone']
                       availableIpdAddressCount = subnet['AvailableIpAddressCount']

                       columns = [ idx_lin, vpcId, cidrBlock, default, lblSubnet, cidrBlockSubnet, az, availableIpdAddressCount ]

                       # Asked to inform if the Subnet is Public? 
                       if subnetIsPublic:
                          resultSubnetIsPublic = "--- "
                          if self._checkSubnetIsPublic(subnet['SubnetId'], None) :
                             resultSubnetIsPublic = "PUBLIC"
                          columns.append(resultSubnetIsPublic)
                       if tableArgs.showTags:
                          tags = Utils.formatPrintTags(vpc['Tags'])
                          columns.append(tags)   

                       prettyTable.addRow(columns)

                   prettyTable.addSeparatorGroup()    
                else:   
                   columns = [ idx_lin, vpcId, cidrBlock, default ]
                   if tableArgs.showTags:
                      tags = Utils.formatPrintTags(vpc['Tags'])
                      columns.append(tags)
                   prettyTable.addRow(columns)

            if (int(tableArgs.sortCol) - 1) > len(columns):
                sortCol = "1"

            prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
            prettyTable.ascendingOrder(not tableArgs.desc)
            result = "VPCs " + "\n\n" + prettyTable.printMe(self.config.tableLineSeparator, tableArgs)
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_VPC]["name"], result, self.config, jsonResult, True, tableArgs)    
        else:
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_VPC]["name"], " No VPC found " + Emoticons.ops(), self.config, jsonResult, True, tableArgs)         

    # List all Ec2 Instances
    def listEc2(self):
       ec2              = self.botoSession().client('ec2')
       instanceId       = None
       privateIPFilter  = None
       vpcIdFilter      = None
       subnetIdFilter   = None
       showSg           = False
       showVpc          = False
       showSubnet       = False
       showImage        = False
       showCpu          = False
       hideName         = False

       tableArgs = TableArgs()
       tableArgs.setArguments(self.config.commandArguments)
       if "," in self.config.commandArguments:
          for arg in self.config.commandArguments.split(","):
              if "i-" in arg:
                 instanceId = arg
              elif arg == "sg":
                 showSg = True
              elif arg == "vpc":
                 showVpc = True    
              elif arg == "subnet":
                 showSubnet = True   
              elif arg == "image":
                 showImage = True 
              elif arg == "cpu":
                 showCpu = True
              elif arg == "noname":
                 hideName = True
              elif arg.startswith("vpc-"):
                 vpcIdFilter = arg
              elif arg.startswith("subnet-"):
                 subnetIdFilter = arg              
              else:
                 # Check IP
                 if Utils.isNumber(arg[0:2]):
                    privateIPFilter = arg

       else:
          arg = self.config.commandArguments
          if "i-" in arg:
            instanceId = arg
          elif arg == "sg":
             showSg = True
          elif arg == "vpc":
             showVpc = True    
          elif arg == "subnet":
             showSubnet = True   
          elif arg == "image":
             showImage = True 
          elif arg == "cpu":
             showCpu = True
          elif arg == "noname": 
             hideName = True
          elif arg.startswith("vpc-"):
             vpcIdFilter = arg
             print(arg)
          elif arg.startswith("subnet-"):
             subnetIdFilter = arg   
          else:
             # Check IP
             if Utils.isNumber(arg[0:2]):
                privateIPFilter = arg


       filters=[]   

       if instanceId:
          filters.append({'Name': 'instance-id','Values': [instanceId]})   
       if privateIPFilter:   
          filters.append({'Name': 'private-ip-address','Values': [privateIPFilter]})
       if vpcIdFilter:   
          filters.append({'Name': 'vpc-id','Values': [vpcIdFilter]})
       if subnetIdFilter:   
          filters.append({'Name': 'subnet-id','Values': [subnetIdFilter]})         

       if self.config.awsTagsToFilter():
           # Environment Tags set
           for tag in self.config.awsTags: 
               filters.append({'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]})   

       # Tags from command line arguments
       if len(tableArgs.tagsTemp) > 0:
          for tag in tableArgs.tagsTemp: 
              filters.append({'Name':'tag:' + tag,'Values':[tableArgs.tagsTemp[tag]]})        

       reservations = ec2.describe_instances( Filters=filters )        

       jsonResult = ""
       if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile: 
          jsonResult = Utils.dictToJson(reservations)

       if reservations and len(reservations["Reservations"]) > 0:
  
          idx_lin = 0

          if hideName:
             header = ["#","Instance Id",       "Type","Launch Time", "Private IP", "Public IP", "State"]
          else:   
             header = ["#","Instance Id","Name","Type","Launch Time", "Private IP", "Public IP", "State"]

          if showSg:             
             header.append("Security Groups")
          if showImage:          
             header.append("Image Id")
          if showCpu:            
             header.append("CPU Options")
          if showVpc:            
             header.append("VPC")
          if showSubnet:         
             header.append("Subnet")
          if tableArgs.showTags: 
             header.append("Tags")

          prettyTable = PrettyTable(header)

          totalListed = 0

          for instanceGroup in reservations["Reservations"]:
              for instance in instanceGroup["Instances"]:
                  idx_lin     += 1
                  totalListed += 1

                  if not hideName:
                     name = ""
                     for t in instance['Tags']:
                         if t["Key"] == "Name":
                            name = t["Value"]

                  launchTimeDate = ""
                  if instance['LaunchTime']:
                     launchTimeDate = instance['LaunchTime'].strftime("%d-%b-%y %H:%M")

                  privateIp = instance['PrivateIpAddress']
                  publicIp  = instance["PublicIpAddress"] if "PublicIpAddress" in instance else ""
                  state     = instance["State"]["Name"]
                  state     = Style.CYAN + state + Style.RESET if state == "stopped" else state

                  if hideName:
                     columns = [ str(idx_lin), str(instance['InstanceId']), instance['InstanceType'], launchTimeDate, privateIp, publicIp, state ]
                  else:   
                     columns = [ str(idx_lin), str(instance['InstanceId']), name, instance['InstanceType'], launchTimeDate, privateIp, publicIp, state ]

                  if showSg:
                     securityGroups = ""
                     for sg in instance["SecurityGroups"]:             
                         if securityGroups != "": securityGroups += ","
                         securityGroups += sg["GroupId"]
                     columns.append(securityGroups)    
                  if showImage:          
                     columns.append(instance["ImageId"])       
                  if showCpu:   
                     cpu  = Style.MAGENTA + "Core:"    + Style.GREEN + str(instance["CpuOptions"]["CoreCount"]) + Style.RESET + " "
                     cpu += Style.MAGENTA + "Threads:" + Style.GREEN + str(instance["CpuOptions"]["ThreadsPerCore"])
                     columns.append(cpu)       
                  if showVpc:            
                     vpc = instance["VpcId"]
                     columns.append(vpc)
                  if showSubnet:         
                     subnet = instance["SubnetId"]
                     columns.append(subnet)
                  if tableArgs.showTags:
                     tags = Utils.formatPrintTags(instance['Tags'])
                     columns.append(tags)

                  prettyTable.addRow(columns)

          resultTxt = ""

          if instanceId:
             if resultTxt != "": resultTxt = resultTxt + "\n"
             resultTxt = resultTxt + "      Instance Id........: " + Style.GREEN + instanceId + Style.RESET
          if privateIPFilter:   
             if resultTxt != "": resultTxt = resultTxt + "\n"
             resultTxt = resultTxt + "      Private IP Address.: " + Style.GREEN + privateIPFilter + Style.RESET
          if vpcIdFilter:   
             if resultTxt != "": resultTxt = resultTxt + "\n"
             resultTxt = resultTxt + "      VPC Id.............: " + Style.GREEN + vpcIdFilter + Style.RESET
          if subnetIdFilter:   
             if resultTxt != "": resultTxt = resultTxt + "\n"
             resultTxt = resultTxt + "      Subnet Id..........: " + Style.GREEN + subnetIdFilter + Style.RESET

          resultTxt = resultTxt + "\n\n Total of EC2s...: " + Style.GREEN + format(totalListed,",") + Style.RESET   

          if (int(tableArgs.sortCol) - 1) > len(columns):
              tableArgs.sortCol = "1"
          prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
          prettyTable.ascendingOrder(not tableArgs.desc)
          result = resultTxt + "\n\n" + prettyTable.printMe(self.config.tableLineSeparator, tableArgs)
          return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_EC2]["name"], result, self.config,jsonResult, True, tableArgs)    
       else:
          resultTxt = ""
          resultTxt = "No EC2 Instance was found. " + Emoticons.ops()
          return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_EC2]["name"], resultTxt, self.config, "", True, tableArgs)

    def botoSession(self):
        return self.config.botoSession()

if __name__ == '__main__':
    config = Config()
    config.awsProfile = "ecomm"
    config.awsRegion = "eu-central-1"
    config.printResults = True
    #config.awsTags["Environment"] = "production"
    config.tableLineSeparator = False
    
    
    # subnets, is-public (only with subnets active)
    # sort1,desc (when requested together with subnets, the group lines will be omitted)
    if len(sys.argv) > 1:
       config.commandArguments = sys.argv[1]
    pyEc2CP = PyEc2CP(config)
    report = pyEc2CP.listEc2()
    Utils.addToClipboard(report)
    print(report)

    # TODO:
    # findEc2 (by IP, by InstanceId, by Tag, etc.)
    # Result show: IP, t3.medium, region, subnet, vpc, security group, (attached Load Balancer optional), (Security Groups Rules optional)




import sys
from pyawscp.Utils import Utils, Style
import locale

class PrettyTable:

    idx_lin            = 0
    content            = []
    columentSorted     = -1
    sizeColumns        = None
    qtdeColumns        = 0
    header             = None
    printedTable       = ""
    addlineSeparator   = True
    groupSeparator     = {}
    sortTablePerformed = False
    ascendingSortOrder = False
    numberSeparator    = False

    def __init__(self, header):
        self.content = []
        self.idx_lin = 0
        self.header  = header

    def isEmpty(self):
        if len(self.content.insert) > 0:
           return False
        return True

    def addRow(self, columns):
        if len(self.header) != len(columns):
            print("HOUSTON, WE HAVE A PROBLEM! The size of Header does not match with this row of columns %d != %d" %( len(self.header), len(columns)) )
            print("Header=" + str(self.header))
            print("Columns=" + str(columns))
            sys.exit()

        # How many columns I will have
        if self.qtdeColumns == 0:
           self.qtdeColumns = len(columns)
        else:
           if len(columns) < self.qtdeColumns or len(columns) > self.qtdeColumns:
               print("HOUSTON, WE HAVE A PROBLEM! This line of columns if invalid, does not have the size of %d" %self.qtdeColumns)
               sys.exit()

        # Define the length of each column by the larger one
        idx_col=0
        # Initiate the Size of Columns if not yet
        marginsLeftRightTotal = 2
        if not self.sizeColumns:
            self.sizeColumns = [0] * len(self.header)
            for header in self.header:
                sizeCol = len(header) + marginsLeftRightTotal
                # In case the columne is smaller than 2, let's put 4
                if sizeCol <= 2:
                    sizeCol += 4
                self.sizeColumns[idx_col] = sizeCol
                idx_col+=1

            # Check the first Line inserted (in case the table has only this one)    
            idx_col=0
            for col in columns:
                # Remove Characters for Colors (do not imply on the size (are invisible))
                col = Utils.removeCharsColors(col)
                addExtraSize = self.checkExtrSizeForNumberFormat(col)    
                if (len(str(col)) + marginsLeftRightTotal + addExtraSize) > (self.sizeColumns[idx_col]):
                   self.sizeColumns[idx_col] = len(str(col)) + marginsLeftRightTotal + addExtraSize
                idx_col+=1    
        else:    
            # If already initiated, iterate and replace only the values with greater length
            for col in columns:
                # Remove Characters for Colors (do not imply on the size (are invisible))
                col = Utils.removeCharsColors(col)
                addExtraSize = self.checkExtrSizeForNumberFormat(col)    
                if (len(str(col)) + marginsLeftRightTotal + addExtraSize) > (self.sizeColumns[idx_col]):
                   self.sizeColumns[idx_col] = len(str(col)) + marginsLeftRightTotal + addExtraSize
                idx_col+=1   

        # Store the values at a new row
        self.content.insert(self.idx_lin, columns)
        self.idx_lin+=1

    def checkExtrSizeForNumberFormat(self,col):
        addExtraSize = 0
        if not self.numberSeparator or not self.is_number(col):
           return addExtraSize

        if int(col) > 999999999:     
            addExtraSize = 3
        elif int(col) > 999999:     
            addExtraSize = 2    
        elif int(col) > 999:
            addExtraSize = 1    
        return addExtraSize    

    def sortByColumn(self, col):
        if self.columentSorted == -1 or self.columentSorted != col:
           self.sortTablePerformed = False

        self.columentSorted = col;

    def addSeparatorGroup(self):
        row = len(self.content) + 1;
        self.groupSeparator[str(row)] = True

    def sortTable(self):
        #Check if the chosen column is numneric of string for the sort
        isSortColumnNumeric = False

        idx_col = 0
        for line in self.content:
            if self.columentSorted == idx_col:
                if self.is_number(line[idx_col]):
                   isSortColumnNumeric = True
                   break

        if isSortColumnNumeric:
           self.content.sort(key=lambda x:int(x[self.columentSorted]), reverse=not self.ascendingSortOrder)
        else:   
           self.content.sort(key=lambda x:x[self.columentSorted], reverse=not self.ascendingSortOrder)

        self.sortTablePerformed = True   

    def printMe(self, addlineSeparator, tableArgs=None):
        self.addlineSeparator = addlineSeparator
        # Implements in the future other formats
        format = "ASCII"

        if not self.sortTablePerformed:
           self.sortTable()

        if format == "ASCII":
           self.printMeInAscii(tableArgs)

        return self.printedTable

    def printHeader(self):
        self.printedTable += self.lineSeparator() + "\n"
        vlr_row = "| "
        idx_col = 0
        for header in self.header:
            number_remain_space = self.sizeColumns[idx_col] - len(header)
            space_remaining     = ' '.ljust(number_remain_space,' ')
            value               = Style.BLUE + header + Style.RESET + space_remaining + Style.RESET
            vlr_row             = vlr_row + value
            idx_col+=1
            vlr_row += "| "    
        self.printedTable += vlr_row + "\n"
        self.printedTable += self.lineSeparator() + "\n"

    def printMeInAscii(self, tableArgs=None):
        self.printHeader()

        #locale.setlocale(locale.LC_ALL, 'es_ES')       
        locale.setlocale(locale.LC_ALL, '')       

        highLightValue_backGroundColor = '\033[103m'

        idx_row = 0
        for row in self.content:
            idx_row       += 1
            vlr_row       = ""
            idx_col       = 0
            for col in row:
                addExtraSize        = self.checkExtrSizeForNumberFormat(col)
                lenCol              = len(Utils.removeCharsColors(str(col))) + addExtraSize
                number_remain_space = self.sizeColumns[idx_col] - lenCol
                space_remaining     = ' '.ljust(number_remain_space,' ')
                value               = Style.GREEN + " " + str(col) + space_remaining + Style.RESET
                if self.is_number(col):
                   decimalPlaces = 0
                   if isinstance(col, (float)):
                      numberCol     = float(col)
                      decimalPlaces = 2
                   else:
                      numberCol = int(col)
                   vlrNumber = Utils.formatNumber(numberCol,3,decimalPlaces,True) if self.numberSeparator and numberCol > 999 else str(col)
                   value = space_remaining + Style.GREEN + vlrNumber + Style.RESET + " "
                vlr_row = vlr_row + "|" + value
                idx_col+=1
            vlr_row += "| "   

            # Mark the Line Number in case there's a highlight --> a " | value " as complement in the command
            originalHighLightValue = self._isThereHighLightValueSeeked(vlr_row, tableArgs)
            if originalHighLightValue:
               vlr_row = Utils.removeCharsColors(vlr_row)                          
               vlr_row = highLightValue_backGroundColor + Style.BLACK + vlr_row.strip() + Style.RESET
               vlr_row = vlr_row.replace(originalHighLightValue, Style.IBLUE + originalHighLightValue + Style.BLACK )

            # Check if filter were requested " | grep value"
            if not tableArgs.filterValue or tableArgs.filterValue.strip() == "":
               # if not, print everything
               self.printedTable += vlr_row + "\n"
            else:
               # otherwise only if if presented on the current line   
               originalFilterValue = self._isThereFilterValueSeekd(vlr_row, tableArgs)
               if originalFilterValue:
                  vlr_row = vlr_row.replace(originalFilterValue, Style.IBLUE + originalFilterValue + Style.GREEN ) 
                  self.printedTable += vlr_row + "\n" 
            
            if self.addlineSeparator:
               self.printedTable += self.lineSeparator() + "\n"
            # In case the Group Separator was asked, close the group when is the case (end of a group, Ex: All the Subnets of a VPC List)
            # If sort was asked, do not add group separator anymore (not sorted approprietaly anymore, group separator does not make sense with sort)   
            if str(idx_row + 1) in self.groupSeparator and (self.columentSorted == 0 and self.ascendingSortOrder):
               self.printedTable += self.lineGroupSeparator() + "\n" 
        
        # Just close the table, even though the line separator wasn't requested
        # Mandatory close even if the option for LineSeparator wasnt required (to close the table, not separate the line)
        # If sort was asked, do not add group separator anymore (not sorted approprietaly anymore, group separator does not make sense with sort)   
        if not self.addlineSeparator and not (str(idx_row + 1) in self.groupSeparator and self.columentSorted == 0 and self.ascendingSortOrder):
           self.printedTable += self.lineSeparator() + "\n"

    def _isThereFilterValueSeekd(self,line,tableArgs):
        if not tableArgs.filterValue or tableArgs.filterValue.strip() == "":
           return None
        originalFilterValue = None   
        filterValueIndex = str(line).lower().find(tableArgs.filterValue.strip().lower())
        if filterValueIndex > -1:
           originalFilterValue = str(line)[filterValueIndex:filterValueIndex+len(tableArgs.filterValue)]
        return originalFilterValue   

    def _isThereHighLightValueSeeked(self, line, tableArgs):
        originalHighLightValue = None
        if not tableArgs.highlightValue or tableArgs.highlightValue.strip() == "":
           return originalHighLightValue
        highLightValueIndex = str(line).lower().find(tableArgs.highlightValue.strip().lower())
        if highLightValueIndex > -1:
           # Grab the original value format (capital letter, small letter format, etc) to colorize below 
           originalHighLightValue = str(line)[highLightValueIndex:highLightValueIndex+len(tableArgs.highlightValue)]
        return originalHighLightValue

    def lineSeparator(self):
        lineSeparator = ""
        for colLength in self.sizeColumns:
            diff = 2
            lineSeparator += '+'.ljust(colLength + diff,'-')
        return lineSeparator + "+"

    def lineGroupSeparator(self):
        lineSeparator = ""
        for colLength in self.sizeColumns:
            diff = 2
            lineSeparator += '+'.ljust(colLength + diff,'=')
        return lineSeparator + "+"     

    def ascendingOrder(self, ascending):
        self.ascendingSortOrder = ascending

    def sortByColumn(self, column):
        if column > self.qtdeColumns:
           print("HOUSTON, WE HAVE A PROBLEM! There's no so much column on your table, check it again")
           sys.exit()
        self.columentSorted = column

    def is_number(self, s):
        try:
            int(s)
            return True
        except ValueError:
            return False        


import logging
import sys, cmd, os, ast, signal
from datetime import datetime
from os.path import expanduser
import configparser
import multiprocessing
from multiprocessing import Process, Pool
import time
import re

# Run temporarily local (python command)
#from pyawscp.utils import utils
from pyawscp.Utils import Utils, Style
from pyawscp.Functions import Functions
from pyawscp.PrettyTable import PrettyTable
from pyawscp.Config import Config
from pyawscp.Emoticons import Emoticons
from pyawscp.PyEc2Cp import PyEc2CP
from pyawscp.PyElbCp import PyElbCP
from pyawscp.PyS3Cp import PyS3CP
from pyawscp.PyR53Cp import PyR53CP
from pyawscp.pymxgraph.PymxGraph import PymxGraph
from pyawscp.pymxgraph.PymxGraphAwsNavigator import PymxGraphAwsNavigator

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)

USER_DIR                   = expanduser("~") + "/.pyawscp/"
FILE_INI                   = USER_DIR + "pyawscp.ini"
LABEL_LENGTH_FUNCTION_NAME = 25
SIZE_SEPARATOR             = 95

PREFERENCES  = "PREFERENCES" 
TRANSFER_S3  = "TRANSFER_S3"

signal.signal(signal.SIGINT, signal.SIG_DFL)

class PyAwsShell(cmd.Cmd):
    Utils.clearScreen()
    intro =  "\n" + \
             Style.CYAN  + "   Wellcome to " + Style.RESET + "AWS" + Style.CYAN + " C:\> " + Style.RESET + "Shell" + Style.CYAN + " $:~" +  Style.RESET + " Cockpit" + Style.CYAN + " version " + Style.RESET + "v0.5" + Style.RESET + "\n\n" + \
             Style.YELLOW  + \
             "      ___        ______   " + Style.CYAN + " ____  _          _ _  " + Style.BLUE + "  ____           _          _ _   " + "\n" + Style.YELLOW + \
             "     / \ \      / / ___|  " + Style.CYAN + "/ ___|| |__   ___| | | " + Style.BLUE + " / ___|___   ___| | ___ __ (_) |_ " + "\n" + Style.YELLOW + \
             "    / _ \ \ /\ / /\___ \  " + Style.CYAN + "\___ \| '_ \ / _ \ | | " + Style.BLUE + "| |   / _ \ / __| |/ / '_ \| | __|" + "\n" + Style.YELLOW + \
             "   / ___ \ V  V /  ___) | " + Style.CYAN + " ___) | | | |  __/ | | " + Style.BLUE + "| |__| (_) | (__|   <| |_) | | |_ " + "\n" + Style.YELLOW + \
             "  /_/   \_\_/\_/  |____/  " + Style.CYAN + "|____/|_| |_|\___|_|_| " + Style.BLUE + " \____\___/ \___|_|\_\ .__/|_|\__|" + "\n" + Style.YELLOW + \
             "                          " + Style.CYAN + "                       " + Style.BLUE + "                     |_|          " + "\n\n" + \
             Style.GREEN + "  Enter " + Style.BLUE + "menu" + Style.GREEN + " or " + Style.BLUE + "ls [-l] " + Style.GREEN + "to list all available AWS functions " + Style.RESET + "\n" + \
             Style.GREEN + "  Enter " + Style.BLUE + "help" + Style.GREEN + " or " + Style.BLUE + "?" + Style.GREEN + "       to list of all commands " + Style.GREEN + "\n" + \
             Style.GREEN + "  Enter " + Style.BLUE + "env" + Style.GREEN + "             to view and setup your environment configuration " + Style.RESET + "\n"  + \
             Style.GREEN + "  Use  "  + Style.BLUE + " <TAB>" + Style.GREEN + "           to autocomplete or view commands/arguments " + Style.RESET + "\n" + \
             Style.GREEN + "  Enter " + Style.BLUE + "navigator" + Style.GREEN + "       to try one of the features... " + Style.CYAN + Emoticons.seeYa() + Style.RESET + "\n"
             
    prompt = '(\033[33maws\033[0m-\033[36mshell\033[0m-\033[34mcockpit\033[0m)\033[33m' + Emoticons.prompt() + '\033[0m:\033[33m~\033[0m '
    ruler      = '-'
    #undoc_header = 'undoc_header'
    doc_header = 'Commands (type help <command>):'
    last_output = ""

    if not os.path.exists(USER_DIR):
       os.makedirs(USER_DIR)

    preferences   = Config()
    configFileIni = None
    if not os.path.exists(FILE_INI):
        # Create the Default INI File
       configFileIni = configparser.ConfigParser(allow_no_value=True)
       configFileIni.add_section(PREFERENCES)
       configFileIni.set(PREFERENCES, "aws-profile","default")
       configFileIni.set(PREFERENCES, "aws-region","eu-central-1")
       configFileIni.set(PREFERENCES, "aws-tags",{})
       configFileIni.set(PREFERENCES, "print-results","false")
       configFileIni.set(PREFERENCES, "table-line-separator","false")

       configFileIni.add_section(TRANSFER_S3)
       configFileIni.set(TRANSFER_S3, "# The size is in Megabytes")
       configFileIni.set(TRANSFER_S3, "chunkSize","10")
       configFileIni.set(TRANSFER_S3, "threshold","100")

       with open(FILE_INI,'w') as configfile:
           configFileIni.write(configfile)
       configFileIni = configparser.ConfigParser(allow_no_value=True)

    configFileIni = configparser.ConfigParser()
    configFileIni.read(FILE_INI)
    preferences.awsProfile               = configFileIni[PREFERENCES]["aws-profile"]
    preferences.awsRegion                = configFileIni[PREFERENCES]["aws-region"]
    preferences.awsTags                  = ast.literal_eval(configFileIni[PREFERENCES]["aws-tags"])
    preferences.printResults             = configFileIni[PREFERENCES]["print-results"] in ['True','true']
    preferences.tableLineSeparator       = configFileIni[PREFERENCES]["table-line-separator"] in ['True','true']
    preferences.uploadChunkSizeMultipart = configFileIni[TRANSFER_S3]["chunkSize"]
    preferences.uploadThresholdMultipart = configFileIni[TRANSFER_S3]["threshold"]

    def do_bye(self, arg):
        "Exit shell\n\n"
        self.close()

    def do_exit(self, arg) :
        "Exit shell\n\n"
        self.close()

    def do_e(self, arg) :
        "[SHORTCUT for \"exit\"]\nQuit shell\n\n"
        self.close()        

    def do_q(self, arg) :
        "[SHORTCUT for \"quit\"]\nQuit shell\n\n"
        self.close()    

    def do_quit(self, arg) :
        "Quit shell\n\n"
        self.close()        

    def do_clear(self,arg):
        "Clear screen\n\n"
        Utils.clearScreen()

    def emptyline(self):
        return cmd.Cmd.emptyline(self)

    def onewayFunction(self, line):
        if not line or len(line.split(",")) < 2:
           return self.default(line)
        programName  = line.split(",")[0]
        functionName = line.split(",")[1]
        args         = line.replace(functionName + ",","").replace(programName + ",","")
        func         = None
        try:
            func = getattr(self, 'do_' + functionName)
            return func(args)    
        except AttributeError:
            return self.default(line.split(",")[1])

    def do_shell(self, line):
        "Run a shell command"
        print ("running shell command:", line)
        output = os.popen(line).read()
        print (output)
        self.last_output = output
    
    def do_echo(self, line):
        "Print the input, replacing '$out' with the output of the last shell command"
        # Obviously not robust
        print (self.last_output)

    def default(self, line):
        LABEL_LENGTH_FUNCTION_NAME = 17
        SIZE_SEPARATOR             = 60
        output  = "-".ljust(SIZE_SEPARATOR,"-") + "\n"           
        output += "   " + Emoticons.tool() + "   " + Style.CYAN + "A W S    C O C K P I T" + Style.RESET + "" + "\n"
        output += "-".ljust(SIZE_SEPARATOR,"-") + "\n"
        output += Style.RESET + "  Function  " + Style.GREEN + line + Style.RESET + \
                  "  do not exist " + Style.CYAN  + "\n  Use the argument \"-h\" or the command \n  \"ls -l\" to see all available options" + Style.RESET + "\n\n"
        self.stdout.write(output)

    def close(self):
        self.savePreferences()
        print(" ")
        print(" See ya! " + Emoticons.seeYa())
        print(" ")
        sys.exit()

    def savePreferences(self):
        if os.path.exists(FILE_INI):
           os.remove(FILE_INI)  

        configFileIni = configparser.ConfigParser()
        configFileIni[PREFERENCES] = {}
        configFileIni[PREFERENCES]["aws-profile"] = self.preferences.awsProfile
        configFileIni[PREFERENCES]["aws-region"] = self.preferences.awsRegion
        configFileIni[PREFERENCES]["aws-tags"] = str(self.preferences.awsTags)
        configFileIni[PREFERENCES]["print-results"] = str(self.preferences.printResults)
        configFileIni[PREFERENCES]["table-line-separator"] = str(self.preferences.tableLineSeparator)

        configFileIni[TRANSFER_S3] = {}
        configFileIni[TRANSFER_S3]["chunksize"] = self.preferences.uploadChunkSizeMultipart
        configFileIni[TRANSFER_S3]["threshold"] = self.preferences.uploadThresholdMultipart

        with open(FILE_INI, "w") as file:
            configFileIni.write(file)
    
    def complete_ad(self,text,line,begidx,endidx):
        return [i for i in ('tags') if i.startswith(text)]
    def do_add(self,arg):
        "Modify environment variables/preferences adding values(type env to see all the available options)\n\n"
        if "tags" in arg:
            tags=arg.replace("tags","")
            if tags:
                for tag in tags.split(","):
                    key = tag.split("=")[0].replace(" ","")
                    vlr = tag.split("=")[1]
                    self.preferences.awsTags[key] = vlr
        else:
            print(" " + Emoticons.error() + " The environment variable/preference " + Style.GREEN + arg.split(" ")[0] + Style.RESET + " is not a collection, use set to chenge its value " + Style.RESET + "\n")

    
    def complete_set(self,text,line,begidx,endidx):
        return [i for i in ('tags','profile','region,''verbose','tables-line','chunk-size','threshold') if i.startswith(text)]

    def do_set(self,arg):
        "Modify environment variables/preferences (type env to see all the available options)\n\n"
        if "tags" in arg:
            tags=arg.replace("tags","")
            if tags:
                self.preferences.awsTags = {}
                for tag in tags.split(","):
                    key = tag.split("=")[0].replace(" ","")
                    vlr = tag.split("=")[1]
                    self.preferences.awsTags[key] = vlr
            else:
               self.preferences.awsTags = {}
        elif len(arg.split(" ")) < 2:
            print(" " + Emoticons.error() + " Ops... Bad command! The syntax is: " + Style.GREEN + "set {VARIABLE} {NEW-VALUE}" + Style.RESET + "\n")
            return None
        elif "profile" in arg:
            value=arg.split(" ")[1]
            self.preferences.awsProfile=value
            print(" " + Emoticons.thumbsUp() + " Profile AWS susccessfully modified to " + Style.GREEN + value + Style.RESET + "\n")
        elif "region" in arg:
            value=arg.split(" ")[1]
            self.preferences.awsRegion=value
            print(" " + Emoticons.thumbsUp() + " Region AWS susccessfully modified to " + Style.GREEN + value + Style.RESET + "\n")     
        elif "verbose" in arg:
            result = (arg.split(" ")[1]).lower() in ("True","true")
            self.preferences.printResults=result
            print(" " + Emoticons.thumbsUp() + " Verbose susccessfully modified to " + Style.GREEN + self.convertsBoolToEnableDisable(result) + Style.RESET + "\n")
        elif "tables-line" in arg:
            result = (arg.split(" ")[1]).lower() in ("True","true")
            self.preferences.tableLineSeparator=result
            print(" " + Emoticons.thumbsUp() + " Tables line susccessfully modified to " + Style.GREEN + self.convertsBoolToEnableDisable(result) + Style.RESET + "\n")
        elif "chunk-size" in arg:
            value=arg.split(" ")[1]    
            self.preferences.uploadChunkSizeMultipart = value
            print(" " + Emoticons.thumbsUp() + " Multipart Upload Chunk-size susccessfully modified to " + Style.GREEN + value + Style.RESET + "\n")
        elif "threshold" in arg:
            value=arg.split(" ")[1]    
            self.preferences.uploadThresholdMultipart = value
            print(" " + Emoticons.thumbsUp() + " Multipart Upload Threshold susccessfully modified to " + Style.GREEN + value + Style.RESET + "\n")    
        else:
            print(" " + Emoticons.error() + " Ops... No preference with the name "+ arg + Style.RESET + "\n")
            return None     

    def complete_get(self,text,line,begidx,endidx):
        return [i for i in ('tags','profile','region','verbose','tables-line','chunk-size','threshold') if i.startswith(text)]
    def do_get(self,arg):        
        "Show a value configured to a environment variable/preference (type env to see all the available options)\n\n"
        if "profile" in arg:
            print(" " + Emoticons.pointRight() + " Profile AWS is " + Style.GREEN + ("default" if self.preferences.awsProfile == "" else self.preferences.awsProfile) + Style.RESET + "\n")
        elif "region" in arg:
            print(" " + Emoticons.pointRight() + " Region AWS is " + Style.GREEN + self.preferences.awsRegion + Style.RESET + "\n")    
        elif "tags" in arg:
            print(" " + Emoticons.pointRight() + " Tags Filters are " + Style.GREEN + str(self.preferences.awsTags) + Style.RESET + "\n")        
        elif "verbose" in arg:
            print(" " + Emoticons.pointRight() + " Verbose is set to " + Style.GREEN + self.convertsBoolToEnableDisable(self.preferences.printResults) + Style.RESET + "\n")    
        elif "tables-line" in arg:
            print(" " + Emoticons.pointRight() + " Tables line is set to " + Style.GREEN + self.convertsBoolToEnableDisable(self.preferences.tableLineSeparator) + Style.RESET + "\n")
        elif "chunk-size" in arg:
            print(" " + Emoticons.pointRight() + " Multipart Upload Chunk-size is set to " + Style.GREEN + self.preferences.uploadChunkSizeMultipart + Style.RESET + "\n")
        elif "threshold" in arg:
            print(" " + Emoticons.pointRight() + " Multipart Upload Threshold is set to " + Style.GREEN + self.preferences.uploadThresholdMultipart + Style.RESET + "\n")        

    def do_env(self, arg):    
        "[SHORTCUT for \"environment\"]\nShow Environment Configuration\n\n"
        self.do_environment(arg)

    def do_environment(self, arg):    
        "Show Environment Configuration\n\n"

        WIDTH_TEXT     = 60
        MARGIN_TEXT    = 20
        SIZE_SEPARATOR = 85
        output  ="-".ljust(SIZE_SEPARATOR,"-") + "\n"           
        
        output += "   " + Emoticons.env() + "   " + Style.CYAN + "E N V I R O N M E N T" + Style.RESET + "" + "\n"           
        output += "-".ljust(SIZE_SEPARATOR,"-") + "\n" 
        output += "  " + Emoticons.pin() + " " + Style.CYAN + "Variables:" + Style.RESET + "" + "\n"
        output +=Style.BLUE + "  AWS Profile.....: " + Style.GREEN + ("default" if self.preferences.awsProfile == "" else self.preferences.awsProfile) + Style.RESET + "\n"
        output +=Style.CYAN + "                    " + "Profile to be used, values set at AWS configuration\n                    file ($USER_HOME/.aws/credentials)" + Style.RESET + "\n"
        output +=Style.BLUE + "                    " + Style.BLUE + "Command to set:" + Style.GREEN+ " set profile {PROFILE}" + Style.RESET + "\n"
        output += "\n"
        output +=Style.BLUE + "  AWS Region......: " + Style.GREEN + self.preferences.awsRegion + Style.RESET + "\n"
        output +=Style.CYAN + "                    " + "AWS Region to be used, values can set at AWS configuration\n                    file ($USER_HOME/.aws/config)" + Style.RESET + "\n"
        output +=Style.BLUE + "                    " + Style.BLUE + "Command to set:" + Style.GREEN+ " set region {PROFILE}" + Style.RESET + "\n"
        output += "\n"
        output +=Style.BLUE + "  Tags............: " + Style.GREEN + str(self.preferences.awsTags) + Style.RESET + "\n"
        output +=Style.CYAN + "                    " + Functions.formatDescr("Tags that's is going to be used as Filters in all AWS Functions (when applicable).",MARGIN_TEXT,WIDTH_TEXT) + Style.RESET + "\n"
        output +=Style.BLUE + "                    " + Style.BLUE + "Command to set:" + Style.GREEN+ " set tags {TAG_NAME=VALUE,TAG_NAME=VALUE}" + Style.RESET + "\n"
        output +=Style.BLUE + "                    " + Style.BLUE + "     or to add:" + Style.GREEN+ " add tags {TAG_NAME=VALUE,TAG_NAME=VALUE}" + Style.RESET + "\n"
        output += "\n"
        output += "\n"
        output += "  " + Emoticons.pin() + " " + Style.CYAN + "Preferences:" + Style.RESET + "" + "\n"
        output +=Style.BLUE + "  Verbose.........: " + Style.GREEN + (self.convertsBoolToEnableDisable(self.preferences.printResults)) + Style.RESET + "\n"
        output +=Style.CYAN + "                    " + "Show any further available information when launch a command." + Style.RESET + "\n"
        output +=Style.BLUE + "                    " + Style.BLUE + "Command to modify:" + Style.GREEN+ " set verbose {true|false}" + Style.RESET + "\n"
        output += "\n"
        output +=Style.BLUE + "  Tables Line.....: " + Style.GREEN + (self.convertsBoolToEnableDisable(self.preferences.tableLineSeparator)) + Style.RESET + "\n"
        output +=Style.CYAN + "                    " + "Remove the lines between table's rows when applied." + Style.RESET + "\n"
        output +=Style.BLUE + "                    " + Style.BLUE + "Command to modify:" + Style.GREEN+ " set tables-line {true|false}" + Style.RESET + "\n"
        output += "\n"
        output += "\n"
        output += "  " + Emoticons.pin() + " " + Style.CYAN + "S3 Multipart Uploads:" + Style.RESET + "" + "\n"
        output +=Style.BLUE + "  Chunk Size......: " + Style.GREEN + self.preferences.uploadChunkSizeMultipart + Style.RESET + "\n"
        output +=Style.CYAN + "                    " + "Upload block size for Multipart Uploads (Megabytes)" + Style.RESET + "\n"
        output +=Style.BLUE + "                    " + Style.BLUE + "Command to modify:" + Style.GREEN+ " set chunk-size 10" + Style.RESET + "\n"
        output +=Style.BLUE + "  Threshold.......: " + Style.GREEN + self.preferences.uploadThresholdMultipart + Style.RESET + "\n"
        output +=Style.CYAN + "                    " + "Level size which starts Multipart Uploads (Megabytes)" + Style.RESET + "\n"
        output +=Style.BLUE + "                    " + Style.BLUE + "Command to modify:" + Style.GREEN+ " set threshold 200" + Style.RESET + "\n"

        output +="-".ljust(SIZE_SEPARATOR,"-") + "\n"                      
        print(output)

    def convertsBoolToEnableDisable(self, value):
        if value:
           return "Enabled"
        return "Disabled"    

    def do_lg(self, arg):
        "List all available groups of functions\n"
        output  = "-".ljust(SIZE_SEPARATOR,"-") + "\n" 
        output += "   " + Emoticons.tool() + "   " + Style.CYAN + "A V A I L A B L E    F U N C T I O N S    G R O U P S" + Style.RESET + "" + "\n"
        output += "-".ljust(SIZE_SEPARATOR,"-") + "\n"    
        for group in Functions.GROUPS:
            output += "  " + Emoticons.pointRight() + " " + Style.GREEN + group + Style.RESET + "\n" #+ \
                      #(" ".ljust(LABEL_LENGTH_FUNCTION_NAME - len(group) + 2,"-")) + "> "+ Functions.formatDescr("description ToDo", 34, 60) + "\n"
        print(output)
        print(Style.GREEN + " To list functions of a group: " + Style.IBLUE + "\n    ls -f networking")
        print(Style.GREEN + "    " + Style.IBLUE + "ls -l -f ec2")
        print("")

    def do_l(self, arg):
        "[SHORTCUT for \"ls -l\"]\nSame as type ls -l, list all available function\n   Arguments:\n   -l  Show usage details of AWS functions\n"
        self.do_ls("-l")    

    def do_ls(self, arg):
        "List all available function\n   Arguments:\n   -l  Show usage details of AWS functions\n"
        output  = "-".ljust(SIZE_SEPARATOR,"-") + "\n" 
        output += "   " + Emoticons.tool() + "   " + Style.CYAN + "A V A I L A B L E    F U N C T I O N S" + Style.RESET + "" + "\n"
        output += "-".ljust(SIZE_SEPARATOR,"-") + "\n"    
        groupPrinted = ""
        i = 0
        for key in Functions.FUNCTIONS:
            i += 1
            function      = Functions.FUNCTIONS[key]
            functionName  = self.convertFunctionsName(key)
            functionDescr = "  " + Emoticons.pointRight() + " " + Style.GREEN + functionName + Style.RESET + (" ".ljust(LABEL_LENGTH_FUNCTION_NAME - len(functionName) + 2,"-")) + \
                                  "> "+ Functions.formatDescr(function["description"], 34, 60) + "\n"
            printIt = True
            if "-f" in arg:
               groupFilter = arg[arg.find("-f") + 3:].split(" ")[0]
               if  function["group"].lower() != groupFilter.lower():
                   printIt = False

            if "-l" in arg:
                if printIt:
                   groupTitle = "  " + Style.IBLUE + function["group"].upper() + Style.RESET + "\n" + "-".ljust(SIZE_SEPARATOR,"-") + "\n"    
                   if groupPrinted != function["group"]:
                      output += groupTitle 
                      groupPrinted = function["group"]
                   output += functionDescr
                   output += Style.MAGENTA + "          Arguments: " + Style.RESET + "\n"
                   for args in function["arguments"]:
                       output += Style.BLUE + "               " + args["name"] + " ".ljust(args["biggerLabel"] - len(args["name"])," ") + Style.MAGENTA 
                       output += ("[Required]" if args["mandatory"] == True else "[Optional]")
                       output += Style.RESET + "....: " + Style.RESET + Functions.formatDescr(args["description"],51,45) + "\n" 
   
                   output += "          " + Style.MAGENTA + "Examples:" + Style.RESET + "\n"
                   for exs in function["examples-interactive"]:
                       output += "               " + exs["command"] + "\n"
                   output += "-".ljust(SIZE_SEPARATOR,"-") + "\n"
            else:
                if printIt:
                   groupTitle    = ("\n" if i > 1 else "") + "  " + Style.IBLUE + function["group"].upper() + Style.RESET + "\n"
                   if groupPrinted != function["group"]:
                      output += groupTitle 
                      groupPrinted = function["group"]
                   output += functionDescr
        print(output)
    def complete_ls(self,text,line,begidx,endidx):
        return ([i for i in ('networking','ec2','s3','general') if i.startswith(text)])    
    
    # List a Help the one function only
    def helpFunction(self, key):    
        function = Functions.FUNCTIONS[key]

        output  = "-".ljust(SIZE_SEPARATOR,"-") + "\n"           
        output += "   " + Emoticons.tool() + " " + Style.CYAN +  function["name"].upper() + Style.RESET + "" + "\n"
        output += "-".ljust(SIZE_SEPARATOR,"-") + "\n"    

        functionName = self.convertFunctionsName(key)
        functionDescr = "  " + Emoticons.pointRight() + " " + Style.GREEN + functionName + Style.RESET + (" ".ljust(LABEL_LENGTH_FUNCTION_NAME - len(functionName) + 2,"-")) + \
                                "> "+ Functions.formatDescr(function["description"], 26, 59) + "\n"
        output += functionDescr
        output += Style.MAGENTA + "          Arguments: " + Style.RESET + "\n"
        for args in function["arguments"]:
            output += Style.BLUE + "               " + args["name"] + " ".ljust(args["biggerLabel"] - len(args["name"])," ") + Style.MAGENTA 
            output += ("[Required]" if args["mandatory"] == True else "[Optional]")
            output += Style.RESET + "....: " + Style.RESET + Functions.formatDescr(args["description"],51,34) + "\n"
        output += "          " + Style.MAGENTA + "Examples:" + Style.RESET + "\n"
        for exs in function["examples-interactive"]:
            output += "               " + exs["command"] + "\n"
        output += "-".ljust(SIZE_SEPARATOR,"-") + "\n"
        return output

    def convertFunctionsName(self,function):
        pos = function.find("-")
        while pos > 1 and pos < 20:
            capitalize = function[pos+1]
            function = function.replace("-"+capitalize,capitalize.upper())
            pos = function.find("-")
        return function


    #########################################################################################################
    # Execute a function, creating a paralell process to not block the UI, 
    # giving some interaction to user while executing a long time function (in case that happens)
    def executeFunction(self,f,arg,preferences=None):
        print (Emoticons.waiting() + " Wait...")
        if self.isWindows():
           p = multiprocessing.Process(target=f,args=(PyAwsShell,arg,preferences,))
           p.start()
           self.waitForCompletion(p)
           p.join()    
        else:    
           p = Process(target=f, args=(arg,preferences,))
           p.start()
           self.waitForCompletion(p)
           p.join()

    def waitForCompletion(self, p):
        tiempo = 0
        idx = 0
        while p.is_alive():
            idx += 1
            if idx == 1:
                sys.stdout.write("\r" + Emoticons.time() + " " + str(int(tiempo)) + " secs " + Emoticons.waitDistract() + "    ")
            elif idx == 2:   
                sys.stdout.write("\r" + Emoticons.time() + " " + str(int(tiempo)) + " secs " + Emoticons.waitDistract() + Emoticons.waitDistract() + "  ")
            elif idx == 3:   
                sys.stdout.write("\r" + Emoticons.time() + " " + str(int(tiempo)) + " secs " + Emoticons.waitDistract() + Emoticons.waitDistract() + Emoticons.waitDistract()) 
            else:   
                sys.stdout.write("\r" + Emoticons.time() + " " + str(int(tiempo)) + " secs   " + Emoticons.waitDistract2() + "  ")  
                idx = 0
            sys.stdout.flush()
            tiempo += 1
            time.sleep(1)

    def isWindows(self):
        if os.name == "nt":
            return True
        return False      

    #########################################################################################################
    def do_subnetIsPublic(self, arg):
        "Check if the subnet is public or not looking for Route Tables with 0.0.0.0/0 route, containing Internet Gateways (igw-*)\n"
        if self.isWindows():
           self.executeFunction(PyAwsShell.subnetIsPublic,arg,self.preferences)
        else:    
           self.executeFunction(self.subnetIsPublic,arg,self.preferences)

    def subnetIsPublic(self, arg, preferences):
        config = preferences
        config.commandArguments = arg.replace("subnet_","subnet-")
        pyEc2Cp = PyEc2CP(config)    
        print(pyEc2Cp.subnetIsPublic())
    def help_subnetIsPublic(self):
        print(self.helpFunction(Functions.SUBNET_IS_PUBLIC))
    def complete_subnetIsPublic(self,text,line,begidx,endidx):
        return ([i for i in ('subnet_074a9ad8389a90bb8','subnet_04863a881ede25af3','subnet_0ab39c4808d8c8c91') if i.startswith(text)])

    #########################################################################################################
    def do_listEc2(self, arg):
        "List EC2 Instances\n"
        if self.isWindows():
           self.executeFunction(PyAwsShell.listEc2,arg,self.preferences)
        else:    
           self.executeFunction(self.listEc2,arg,self.preferences)
    def listEc2(self, arg, preferences):
        config = preferences
        config.commandArguments = arg.replace("i_","i-")
        pyEc2Cp = PyEc2CP(config)
        report  = pyEc2Cp.listEc2()
        Utils.addToClipboard(report)
        self.last_output = report
        print(report)
    def help_listEc2(self):
        print(self.helpFunction(Functions.LIST_EC2))            
    def complete_listEc2(self,text,line,begidx,endidx):
        return ([i for i in ('i_0902ce6ab97baa8e7','i_0bbb04a9cd37a4675','10.2.1.0','172.0.1.123','sg','vpc','subnet','image','cpu','noname','showtags','desc','tags','sort3','tags:Environment=production#Project=xpto') if i.startswith(text)])

    #########################################################################################################
    def do_listSubnetsVpc(self, arg):
        "List all the Subnets of a VPC\n"
        if self.isWindows():
           self.executeFunction(PyAwsShell.listSubnetsVpc,arg,self.preferences)
        else:    
           self.executeFunction(self.listSubnetsVpc,arg,self.preferences)
    def listSubnetsVpc(self, arg, preferences):
        config = preferences
        config.commandArguments = arg.replace("vpc_","vpc-")
        pyEc2Cp = PyEc2CP(config)
        report  = pyEc2Cp.listSubnetsVpc()
        Utils.addToClipboard(report)
        self.last_output = report
        print(report)
    def help_listSubnetsVpc(self):
        print(self.helpFunction(Functions.LIST_SUBNETS_VPC))            
    def complete_listSubnetsVpc(self,text,line,begidx,endidx):
        return ([i for i in ('vpc_01c6810fb391d215b','vpc_0834b01b2cecac1f3','ispublic','showtags','desc','tags','sort3') if i.startswith(text)])        

    #########################################################################################################
    def do_listVpc(self, arg):
        "List all the VPCs\n"
        if self.isWindows():
           self.executeFunction(PyAwsShell.listVpc,arg,self.preferences)
        else:    
           self.executeFunction(self.listVpc,arg,self.preferences)
    def listVpc(self, arg, preferences):
        config = preferences
        config.commandArguments = arg.replace("is_","is")
        pyEc2Cp = PyEc2CP(config)
        report = pyEc2Cp.listVpc()
        Utils.addToClipboard(report)
        self.last_output = report
        print(report)
    def help_listVpc(self):
        print(self.helpFunction(Functions.LIST_VPC))    
    def complete_listVpc(self,text,line,begidx,endidx):
        return ([i for i in ('subnets','ispublic','showtags','desc','tags','sort3') if i.startswith(text)])        

    #########################################################################################################
    def do_listTargetGroups(self, arg):
        "List all the Target Groups\n"
        if self.isWindows():
           self.executeFunction(PyAwsShell.listTargetGroups,arg,self.preferences)
        else:   
           self.executeFunction(self.listTargetGroups,arg,self.preferences)
    def listTargetGroups(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyElbCP = PyElbCP(config)
        report = pyElbCP.listTargetGroups()
        Utils.addToClipboard(report)
        self.last_output = report
        print(report)
    def help_listTargetGroups(self):
        print(self.helpFunction(Functions.LIST_TARGET_GROUPS))    
    def complete_listTargetGroups(self,text,line,begidx,endidx):
        return ([i for i in ('health','sort3','desc') if i.startswith(text)])        

    #########################################################################################################
    def do_findElbsEc2(self, arg):
        "Find all the Load Balancer and Target Groups pointing to an EC2 Instance\n"
        if self.isWindows():
           self.executeFunction(PyAwsShell.findElbsEc2,arg,self.preferences)
        else:
           self.executeFunction(self.findElbsEc2,arg,self.preferences)    
    def findElbsEc2(self, arg, preferences):
        config = preferences
        config.commandArguments = arg.replace("i_","i-")
        pyElbCP = PyElbCP(config)
        report = pyElbCP.findInstancesElbsTgs()
        Utils.addToClipboard(report)
        self.last_output = report
        print(report)
    def help_findElbsEc2(self):
        print(self.helpFunction(Functions.FIND_ELBS_EC2))            
    def complete_findElbsEc2(self,text,line,begidx,endidx):
        return ([i for i in ('i_0361f6405c08923b6','i_03ef7e40cc1ba806f','i_046adc50a97a02924','elbarn','tgarn','sort3','desc') if i.startswith(text)])        
    
    #########################################################################################################
    def do_listBucketsS3(self, arg):
        "List buckets of S3\n"
        if self.isWindows():
           self.executeFunction(PyAwsShell.listBucketsS3,arg,self.preferences)
        else:
           self.executeFunction(self.listBucketsS3,arg,self.preferences)    
    def listBucketsS3(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyS3CP = PyS3CP(config)
        report = pyS3CP.listBucketsS3()
        Utils.addToClipboard(report)
        self.last_output = report
        print(report)
    def help_listBucketsS3(self):
        print(self.helpFunction(Functions.LIST_BUCKETS_S3))            

    #########################################################################################################
    def do_listObjectsBucketS3(self, arg):
        "List objects in a bucket\n"
        if self.isWindows():
           self.executeFunction(PyAwsShell.listObjectsBucketS3,arg,self.preferences)
        else:
           self.executeFunction(self.listObjectsBucketS3,arg,self.preferences)    
    def listObjectsBucketS3(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyS3CP = PyS3CP(config)
        report = pyS3CP.listObjectsBucketS3()
        Utils.addToClipboard(report)
        self.last_output = report
        print(report)
    def help_listObjectsBucketS3(self):
        print(self.helpFunction(Functions.LIST_OBJECTS_BUCKETS_S3))            
    def complete_listObjectsBucketS3(self,text,line,begidx,endidx):
        return ([i for i in ('prefix=') if i.startswith(text)])    

    #########################################################################################################
    def do_listUploadsS3(self, arg):
        "List Uploads that did not finished yet (know as MultiPart Uploads)\n"
        self.listUploadsS3(arg,self.preferences)
    def listUploadsS3(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyS3CP = PyS3CP(config)
        report = pyS3CP.listUploads()
        Utils.addToClipboard(report)
        self.last_output = report
        print(report)
    def help_listUploadsS3(self):
        print(self.helpFunction(Functions.LIST_UPLOADS_S3))            

    #########################################################################################################
    def do_abortUploadsS3(self, arg):
        "Abort Uploads that did not finished yet (know as MultiPart Uploads)\n"
        self.abortUploadsS3(arg,self.preferences)
    def abortUploadsS3(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyS3CP = PyS3CP(config)
        report = pyS3CP.abortUploads()
        Utils.addToClipboard(report)
        self.last_output = report
        print(report)
    def help_abortUploadsS3(self):
        print(self.helpFunction(Functions.ABORT_UPLOADS_S3))

    #########################################################################################################
    def do_transferToS3(self, arg):
        "Upload file/folder contents to a S3 Bucket\n"
        self.transferToS3(arg,self.preferences)
    def transferToS3(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyS3CP = PyS3CP(config)
        report = pyS3CP.transferToS3()
        if report and report != "":
           print(report)
    def help_transferToS3(self):
        print(self.helpFunction(Functions.TRANSFER_TO_S3)) 

    #########################################################################################################
    def do_showTargetsS3(self, arg):
        "Show all the targets used for Uploads (files, folders)\n"
        self.showTargetsS3(arg,self.preferences)
    def showTargetsS3(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyS3CP = PyS3CP(config)
        pyS3CP.showTargetsS3()
    def help_showTargetsS3(self):
        print(self.helpFunction(Functions.SHOW_TARGET_S3))

    #########################################################################################################
    def do_showTargetUploadsS3(self, arg):
        "List all uploads performed to a specific target (files/folder)\n"
        self.showTargetUploadsS3(arg,self.preferences)
    def showTargetUploadsS3(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyS3CP = PyS3CP(config)
        pyS3CP.showTargetUploadsS3()
    def help_showTargetUploadsS3(self):
        print(self.helpFunction(Functions.SHOW_TARGET_UPLOADS_S3))

    #########################################################################################################
    def do_showTargetUploadPartsS3(self, arg):
        "List all uploads performed to a specific target (files/folder)\n"
        self.showTargetUploadPartsS3(arg,self.preferences)
    def showTargetUploadPartsS3(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyS3CP = PyS3CP(config)
        pyS3CP.showTargetUploadPartsS3()
    def help_showTargetUploadPartsS3(self):
        print(self.helpFunction(Functions.SHOW_TARGET_UPLOAD_PARTS_S3))                            

    #########################################################################################################
    def do_removeTargetS3(self, arg):
        "Delete all the data of a target (File/folder) upload\n"
        self.removeTargetS3(arg,self.preferences)
    def removeTargetS3(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyS3CP = PyS3CP(config)
        pyS3CP.removeTargetS3()
    def help_removeTargetS3(self):
        print(self.helpFunction(Functions.REMOVE_TARGET_HISTORY_S3))                                

    #########################################################################################################
    
    def do_nslookupEc2R53(self, arg):
        "Query and describe the route path from a DNS domain until a EC2 Instance\n"
        if self.isWindows():
           self.executeFunction(PyAwsShell.nslookupEc2R53,arg,self.preferences)  
        else:    
           self.executeFunction(self.nslookupEc2R53,arg,self.preferences)
    def nslookupEc2R53(self, arg, preferences):
        config = preferences
        config.commandArguments = arg
        pyr53CP = PyR53CP(config)
        report = pyr53CP.nslookupEc2R53()
        Utils.addToClipboard(report)
        self.last_output = report
        print(report)
    def help_nslookupEc2R53(self):
        print(self.helpFunction(Functions.NSLOOKUP_EC2_R53))            
    def complete_nslookupEc2R53(self,text,line,begidx,endidx):
        return ([i for i in ('graph','thin','save','verbose') if i.startswith(text)])

    #########################################################################################################
    def do_navigator(self, arg):
        "Navigate through the Network information of the current AWS Account\n"
        self.navigator(arg,self.preferences)
    def navigator(self, arg, preferences):
        verbose = False
        config  = preferences
        config.commandArguments = arg
        if "verbose" in config.commandArguments:
            verbose = True
        #pathToResources = "./pymxgraph"
        pathToResources = "./"
        pymxGraph = PymxGraph(pathToResources)
        pymxGraphAwsNavigator = PymxGraphAwsNavigator(pathToResources, pymxGraph.images, pymxGraph.htmlSnippets, verbose)
        pymxGraphAwsNavigator.drawAwsNavigator()
    def help_navigator(self):
        print(self.helpFunction(Functions.AWS_NAVIGATOR))            
    def complete_navigator(self,text,line,begidx,endidx):
        return ([i for i in ('verbose','') if i.startswith(text)])    




#from pyawscp.Utils import Style
from .Utils import Style

class Functions:

    SUBNET_IS_PUBLIC             = "subnetIsPublic"
    LIST_SUBNETS_VPC             = "listSubnetsVpc"
    LIST_VPC                     = "listVpc"
    LIST_TARGET_GROUPS           = "listTargetGroups"
    FIND_ELBS_EC2                = "findElbsEc2"
    LIST_EC2                     = "listEc2"
    LIST_BUCKETS_S3              = "listBucketsS3"
    LIST_OBJECTS_BUCKETS_S3      = "listObjectsBucketS3"
    LIST_UPLOADS_S3              = "listUploadsS3"
    ABORT_UPLOADS_S3             = "abortUploadsS3"
    TRANSFER_TO_S3               = "transferToS3"
    SHOW_TARGET_S3               = "showTargetsS3"
    SHOW_TARGET_UPLOADS_S3       = "showTargetUploadsS3"
    SHOW_TARGET_UPLOAD_PARTS_S3  = "showTargetUploadPartsS3"
    REMOVE_TARGET_HISTORY_S3     = "removeTargetHistoryS3"
    NSLOOKUP_EC2_R53             = "nslookupEc2R53"
    AWS_NAVIGATOR                = "navigator"
   
    verboseDescription           = "Show information in JSON format."
    saveDescription              = "Save the results to a JSON file."
    highLightDescription         = "Hightlight content using \"| {VALUE}\""
    filterDescription            = "Filter content using \"| grep {VALUE}\""

    NETWORKING = "Networking"
    EC2        = "EC2"
    S3         = "S3"
    GENERAL    = "General"
    
    GROUPS = [NETWORKING, EC2, S3, GENERAL]

    FUNCTIONS = {
        SUBNET_IS_PUBLIC:{
            "group": NETWORKING,
            "name": "Subnet is Public",
            "description": "Check if the subnet is public looking for Route Tables with route 0.0.0.0/0 containing Internet Gateways (igw-*)",
            "arguments": [
                {"name": "{SUBNET_ID}","mandatory": True, "description": "Id of the Subnet","biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "subnetIsPublic " + Style.BLUE + "subnet-074a9ad8389a90bb8" + Style.RESET}
            ]
        },
        LIST_SUBNETS_VPC:{
            "group": NETWORKING,
            "name": "List Subnets VPC",
            "description": "List all Subnets of a VPC",
            "arguments": [
                {"name": "{VPC_ID}",           "mandatory": True,  "description": "Id of the VPC","biggerLabel": 20},
                {"name": "ispublic",           "mandatory": False, "description": "Add a column info showing if the subnet is Public","biggerLabel": 20},
                {"name": "sort{Column Number}","mandatory": False, "description": "Column number to sort the table result","biggerLabel": 20},
                {"name": "desc",               "mandatory": False, "description": "Do a desceding sort","biggerLabel": 20},
                {"name": "tags",               "mandatory": False, "description": "Tags to filter the resource (will be added to the Environment Tags set)","biggerLabel": 20},
                {"name": "showtags",           "mandatory": False, "description": "Add a column with all Tags of the Subnet.","biggerLabel": 20},
                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20},
                {"name": "save",               "mandatory": False, "description": saveDescription,"biggerLabel": 20},
                {"name": "| {VALUE}",          "mandatory": False, "description": highLightDescription,"biggerLabel": 20},
                {"name": "| grep {VALUE}",     "mandatory": False, "description": filterDescription,"biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "listSubnetsVpc " + Style.BLUE + "vpc-050a1a1dccd178b9a" + Style.RESET},
                {"command":Style.GREEN + "listSubnetsVpc " + Style.BLUE + "vpc-050a1a1dccd178b9a,tags:Environment=production#Project=xpto" + Style.RESET},
                {"command":Style.GREEN + "listSubnetsVpc " + Style.BLUE + "vpc-050a1a1dccd178b9a,ispublic" + Style.RESET},
                {"command":Style.GREEN + "listSubnetsVpc " + Style.BLUE + "vpc-050a1a1dccd178b9a,sort3,desc" + Style.RESET},
                {"command":Style.GREEN + "listSubnetsVpc " + Style.BLUE + "vpc-050a1a1dccd178b9a,tag:Environment=production,sort3" + Style.RESET},
                {"command":Style.GREEN + "listSubnetsVpc " + Style.BLUE + "vpc-050a1a1dccd178b9a,ispublic" + Style.RESET},
                {"command":Style.GREEN + "listSubnetsVpc " + Style.BLUE + "vpc-050a1a1dccd178b9a,save" + Style.RESET}
            ]
        },
        LIST_VPC:{
            "group": NETWORKING,
            "name": "List VPCs",
            "description": "List all VPCs (and its subnets)",
            "arguments": [
                {"name": "subnets",            "mandatory": False, "description": "List all the subnets of the VPCs, grouping by VPC Id. Notice: the grouping will be disabled when sort/desc is requested.","biggerLabel": 20},
                {"name": "ispublic",           "mandatory": False, "description": "To use in conjunction with subnets, adding a column saying if the Subnet is public.","biggerLabel": 20},
                {"name": "sort{Column Number}","mandatory": False, "description": "Column number to sort the table result","biggerLabel": 20},
                {"name": "desc",               "mandatory": False, "description": "Do a desceding sort","biggerLabel": 20},
                {"name": "tags",               "mandatory": False, "description": "Tags to filter the resource (will be added to the Environment Tags set)","biggerLabel": 20},
                {"name": "showtags",           "mandatory": False, "description": "Add a column with all Tags of the VPC.","biggerLabel": 20},
                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20},
                {"name": "save",               "mandatory": False, "description": saveDescription,"biggerLabel": 20},
                {"name": "| {VALUE}",          "mandatory": False, "description": highLightDescription,"biggerLabel": 20},
                {"name": "| grep {VALUE}",     "mandatory": False, "description": filterDescription,"biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "listVpc " + Style.RESET},
                {"command":Style.GREEN + "listVpc " + Style.BLUE + "tags:Environment=production#Project=xpto" + Style.RESET},
                {"command":Style.GREEN + "listVpc " + Style.BLUE + "subnets" + Style.RESET},
                {"command":Style.GREEN + "listVpc " + Style.BLUE + "subnets,ispublic" + Style.RESET},
                {"command":Style.GREEN + "listVpc " + Style.BLUE + "subnets,ispublic,showtags" + Style.RESET},
                {"command":Style.GREEN + "listVpc " + Style.BLUE + "sort1" + Style.RESET},
                {"command":Style.GREEN + "listVpc " + Style.BLUE + "sort3,desc" + Style.RESET}
            ]
        },
        LIST_TARGET_GROUPS:{
            "group": NETWORKING,
            "name": "List Target Groups",
            "description": "List all the Target Groups",
            "arguments": [
                {"name": "Target Groups ARN",  "mandatory": False, "description": "List only the requested Target Groups, to pass more than one use \";\" to separate them. This argument is exclusive with Load Balancer option, it will overwrite it if both were chosen.","biggerLabel": 20},
                {"name": "Load Balancer ARN",  "mandatory": False, "description": "List all the Target Groups associated with the requested Load Balancer.","biggerLabel": 20},
                {"name": "health",             "mandatory": False, "description": "Add the columns of Health Check. Notice! This might exceed the horizontal width of the Terminal Window.","biggerLabel": 20},
                {"name": "tgarn",              "mandatory": False, "description": "Show the Target Groups ARN, otherwise will show only the name.","biggerLabel": 20},
                {"name": "elbarn",             "mandatory": False, "description": "Add a column with the complete ELB ARN, other just the name will be presented.","biggerLabel": 20},
                {"name": "sort{Column Number}","mandatory": False, "description": "Column number to sort the table result","biggerLabel": 20},
                {"name": "desc",               "mandatory": False, "description": "Do a desceding sort","biggerLabel": 20},
                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20},
                {"name": "save",               "mandatory": False, "description": saveDescription,"biggerLabel": 20},
                {"name": "| {VALUE}",          "mandatory": False, "description": highLightDescription,"biggerLabel": 20},
                {"name": "| grep {VALUE}",     "mandatory": False, "description": filterDescription,"biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "listTargetGroups " + Style.RESET},
                {"command":Style.GREEN + "listTargetGroups " + Style.BLUE + " arn:aws:elb:xxx:123456:loadbalancer/app/LB-Gandalf" + Style.RESET},
                {"command":Style.GREEN + "listTargetGroups " + Style.BLUE + " arn:aws:elb:xxx:123456:targetgroup/app/TG-Frodo" + Style.RESET},
                {"command":Style.GREEN + "listTargetGroups " + Style.BLUE + " arn:aws:elb:xxx:123456:targetgroup/app/TG-Frodo;\n                                 arn:aws:elb:xxx:123456:targetgroup/app/TG-Sam" + Style.RESET},
                {"command":Style.GREEN + "listTargetGroups " + Style.BLUE + " arn:aws:elb:xxx:123456:loadbalancer/app/LB-Gandal,\n                                 health" + Style.RESET},
                {"command":Style.GREEN + "listTargetGroups " + Style.BLUE + " arn:aws:elb:xxx:123456:loadbalancer/app/LB-Gandal,\n                                 health,sort1" + Style.RESET},
                {"command":Style.GREEN + "listTargetGroups " + Style.BLUE + " sort3,desc" + Style.RESET},
                {"command":Style.GREEN + "listTargetGroups " + Style.BLUE + " | production" + Style.RESET}
            ]
        },
        FIND_ELBS_EC2:{
            "group": NETWORKING,
            "name": "Find ELBs targeting/pointing to a specific EC2 Instance",
            "description": "Find Load Balancers and Target Groups pointing to a specific EC2 Instance",
            "arguments": [
                {"name": "EC2 Instance Id",    "mandatory": True,  "description": "EC2 Instance Id that the Load Balancers are pointing to.","biggerLabel": 20},
                {"name": "elbarn",             "mandatory": False, "description": "Shoe the Load Balance ARN, otherwise will show the name.","biggerLabel": 20},
                {"name": "tgarn",              "mandatory": False, "description": "Show the Target Groups ARN, otherwise will show the name.","biggerLabel": 20},
                {"name": "sort{Column Number}","mandatory": False, "description": "Column number to sort the table result","biggerLabel": 20},
                {"name": "desc",               "mandatory": False, "description": "Do a desceding sort","biggerLabel": 20},
                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20},
                {"name": "save",               "mandatory": False, "description": saveDescription,"biggerLabel": 20},
                {"name": "| {VALUE}",          "mandatory": False, "description": highLightDescription,"biggerLabel": 20},
                {"name": "| grep {VALUE}",     "mandatory": False, "description": filterDescription,"biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "findElbsEc2 " + Style.RESET},
                {"command":Style.GREEN + "findElbsEc2 " + Style.BLUE + " i-0361f6405c08923b6" + Style.RESET},
                {"command":Style.GREEN + "findElbsEc2 " + Style.BLUE + " i-0361f6405c08923b6,elbarn" + Style.RESET},
                {"command":Style.GREEN + "findElbsEc2 " + Style.BLUE + " i-0361f6405c08923b6,sort3,desc" + Style.RESET}
            ]
        },
        NSLOOKUP_EC2_R53:{
            "group": NETWORKING,
            "name": "Describe route from a DNS Domain to EC2 Instances",
            "description": "From a DNS domain, since Route 53, it will query and describe the route path all the day down to an EC2 Instance/IP",
            "arguments": [
                {"name": "{DNS}",              "mandatory": True,  "description": "DNS to describe the path","biggerLabel": 20},
                {"name": "thin",               "mandatory": False, "description": "Present less information, a summary form","biggerLabel": 20},
                {"name": "graph",              "mandatory": False, "description": "Generates a graph and opens in a browser","biggerLabel": 20},
                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20},
                {"name": "save",               "mandatory": False, "description": saveDescription,"biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "nslookupEc2R53 " + Style.BLUE + " prd.jenkins.mycompany.com" + Style.RESET},
                {"command":Style.GREEN + "nslookupEc2R53 " + Style.BLUE + " prd.jenkins.mycompany.com,thin" + Style.RESET},
                {"command":Style.GREEN + "nslookupEc2R53 " + Style.BLUE + " prd.jenkins.mycompany.com,thin,save" + Style.RESET},
                {"command":Style.GREEN + "nslookupEc2R53 " + Style.BLUE + " prd.jenkins.mycompany.com,save,verbose" + Style.RESET}
            ]
        },
        LIST_EC2:{
            "group": EC2,
            "name": "List EC2 Instances",
            "description": "List EC2s Instances",
            "arguments": [
                {"name": "{INSTANCE ID}",      "mandatory": False, "description": "Filter with theInstance Id  requested, otherwise all will be listed","biggerLabel": 20},
                {"name": "{PRIVATE IP}",       "mandatory": False, "description": "Filter with the Private IP requested, otherwise all will be listed","biggerLabel": 20},
                {"name": "{VPC ID}",           "mandatory": False, "description": "Filter with the VPC Id requested, otherwise all will be listed","biggerLabel": 20},
                {"name": "{SUBNET ID}",        "mandatory": False, "description": "Filter with the Subnet Id requested, otherwise all will be listed","biggerLabel": 20},

                {"name": "sg",                 "mandatory": False, "description": "Add a column with the Security Groups associated to the EC2 Instance","biggerLabel": 20},
                {"name": "vpc",                "mandatory": False, "description": "Add a column with the VPC Id where the EC2 Instance is located","biggerLabel": 20},
                {"name": "subnet",             "mandatory": False, "description": "Add a column with the Subnet Id where the EC2 Instance is located","biggerLabel": 20},
                {"name": "image",              "mandatory": False, "description": "Add a column with the Image Id of the EC2 Instance","biggerLabel": 20},
                {"name": "cpu",                "mandatory": False, "description": "Add a column with the information of EC2 Instance CPU Options","biggerLabel": 20},
                {"name": "noname",             "mandatory": False, "description": "Remove the column name information of EC2 Instance","biggerLabel": 20},

                {"name": "sort{Column Number}","mandatory": False, "description": "Column number to sort the table result","biggerLabel": 20},
                {"name": "desc",               "mandatory": False, "description": "Do a desceding sort","biggerLabel": 20},
                {"name": "tags",               "mandatory": False, "description": "Tags to filter the resource (will be added to the Environment Tags set)","biggerLabel": 20},
                {"name": "showtags",           "mandatory": False, "description": "Add a column with all Tags of the VPC.","biggerLabel": 20},
                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20},
                {"name": "save",               "mandatory": False, "description": saveDescription,"biggerLabel": 20},
                {"name": "| {VALUE}",          "mandatory": False, "description": highLightDescription,"biggerLabel": 20},
                {"name": "| grep {VALUE}",     "mandatory": False, "description": filterDescription,"biggerLabel": 20}

            ],
            "examples-interactive": [
                {"command":Style.GREEN + "listEc2 " + Style.RESET},
                {"command":Style.GREEN + "listEc2 " + Style.BLUE + "i-0902ce6ab97baa8e7" + Style.RESET},
                {"command":Style.GREEN + "listEc2 " + Style.BLUE + "i-0bbb04a9cd37a4675,showtags" + Style.RESET},
                {"command":Style.GREEN + "listEc2 " + Style.BLUE + "10.115.201.1" + Style.RESET},
                {"command":Style.GREEN + "listEc2 " + Style.BLUE + "172.168.239.33,showtags" + Style.RESET},
                {"command":Style.GREEN + "listEc2 " + Style.BLUE + "tags:Environment=production#Project=xpto" + Style.RESET},
                {"command":Style.GREEN + "listEc2 " + Style.BLUE + "showtags" + Style.RESET},
                {"command":Style.GREEN + "listEc2 " + Style.BLUE + "sort1" + Style.RESET},
                {"command":Style.GREEN + "listEc2 " + Style.BLUE + "sort3,desc" + Style.RESET}
            ]
        },
        LIST_UPLOADS_S3:{
            "group": S3,
            "name": "List Uploads that did not finished yet",
            "description": "List Uploads that did not finished yet, known as MultiPart Uploads",
            "arguments": [
                {"name": "{BUCKET}",           "mandatory": True,  "description": "Name of the bucket","biggerLabel": 20},
                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20},
                {"name": "save",               "mandatory": False, "description": saveDescription,"biggerLabel": 20},
                {"name": "| {VALUE}",          "mandatory": False, "description": highLightDescription,"biggerLabel": 20},
                {"name": "| grep {VALUE}",     "mandatory": False, "description": filterDescription,"biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "listUploadsS3 " + Style.BLUE + " bucketName" + Style.RESET},
                {"command":Style.GREEN + "listUploadsS3 " + Style.BLUE + " bucketName/key1/" + Style.RESET},
                {"command":Style.GREEN + "listUploadsS3 " + Style.BLUE + " bucketName/key1/key2/" + Style.RESET},
                
            ]
        },
        ABORT_UPLOADS_S3:{
            "group": S3,
            "name": "Abort Uploads that did not finished yet",
            "description": "Abort Uploads that did not finished yet, known as MultiPart Uploads",
            "arguments": [
                {"name": "{BUCKET}",           "mandatory": True,  "description": "Name of the bucket","biggerLabel": 20},
                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20},
                {"name": "save",               "mandatory": False, "description": saveDescription,"biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "abortUploadsS3 " + Style.BLUE + " bucketName" + Style.RESET}
            ]
        },
        TRANSFER_TO_S3:{
            "group": S3,
            "name": "Upload files/folder to a Bucket S3",
            "description": "Upload files/folder to a Bucket S3 using the MultiPart Uploads (can be resumed in case drop in the middle of the process)",
            "arguments": [
                {"name": "{BUCKET}",           "mandatory": True, "description": "Name of the bucket to receive the content","biggerLabel": 20},
                {"name": "{FILE|FOLDER}",      "mandatory": True, "description": "The file or folder to be uploaded","biggerLabel": 20},
                {"name": "recursive",          "mandatory": False,"description": "In case a folder, the copy will be recursive (all subfolders)","biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "transferToS3 " + Style.BLUE + " bucketName,c:\\MyFiles\\logs-armagedon.tar.gz" + Style.RESET},
                {"command":Style.GREEN + "transferToS3 " + Style.BLUE + " bucketName,c:\\MyFiles" + Style.RESET},
                {"command":Style.GREEN + "transferToS3 " + Style.BLUE + " bucketName,c:\\MyFiles,recursive" + Style.RESET},
                {"command":Style.GREEN + "transferToS3 " + Style.BLUE + " iron-maiden,~/Downloads/repo-com.tar.gz" + Style.RESET},
                {"command":Style.GREEN + "transferToS3 " + Style.BLUE + " iron-maiden,/home/ualter/Downloads/repo-com.tar.gz" + Style.RESET},
                {"command":Style.GREEN + "transferToS3 " + Style.BLUE + " iron-maiden,/home/ualter/Downloads,recursive" + Style.RESET}
            ]
        },
        SHOW_TARGET_S3:{
            "group": S3,
            "name": "Show all the targets used for Uploads (files, folders) to S3",
            "description": "Show all history of files/folders (Targets) you have used to upload to S3",
            "arguments": [
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "showTargetsS3 " + Style.RESET}
            ]
        },
        SHOW_TARGET_UPLOADS_S3:{
            "group": S3,
            "name": "Show all uploads history of a target",
            "description": "List all uploads performed to a specific target (files/folder)",
            "arguments": [
                {"name": "{TARGET}",           "mandatory": True, "description": "Name of the target (File or Folder)","biggerLabel": 20},
                {"name": "showAll",            "mandatory": False,"description": "Verbose, list all the columns available","biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "showTargetUploadsS3 " + Style.BLUE + "d:\\Downloads\\repository.zip" + Style.RESET},
                {"command":Style.GREEN + "showTargetUploadsS3 " + Style.BLUE + "d:\\Downloads\\repository.zip,showAll" + Style.RESET},
                {"command":Style.GREEN + "showTargetUploadsS3 " + Style.BLUE + "d:\\MyFiles\\" + Style.RESET}
            ]
        },
        SHOW_TARGET_UPLOAD_PARTS_S3:{
            "group": S3,
            "name": "Show all parts of a Multipart Upload",
            "description": "List all parts of a performed Multipart Upload",
            "arguments": [
                {"name": "{TARGET}",           "mandatory": True, "description": "Name of the target (File or Folder)","biggerLabel": 20},
                {"name": "key",                "mandatory": True,"description": "Name of the file/key","biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "showTargetUploadPartsS3 " + Style.BLUE + "d:\\Downloads\\repository.zip,repository.zip" + Style.RESET},
                {"command":Style.GREEN + "showTargetUploadPartsS3 " + Style.BLUE + "d:\\Downloads\\,myFile.tar.gz" + Style.RESET}
            ]
        },
        REMOVE_TARGET_HISTORY_S3:{
            "group": S3,
            "name": "Delete the history of a target upload",
            "description": "Delete all the data of a target (File/folder) upload",
            "arguments": [
                {"name": "{TARGET}",           "mandatory": True, "description": "Name of the target (File or Folder)","biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "removeTargetHistoryS3 " + Style.BLUE + "d:\\Downloads\\repository.zip" + Style.RESET},
                {"command":Style.GREEN + "removeTargetHistoryS3 " + Style.BLUE + "d:\\Downloads\\" + Style.RESET}
            ]
        },
        LIST_BUCKETS_S3:{
            "group": S3,
            "name": "List all buckets",
            "description": "List all buckets",
            "arguments": [
                {"name": "csv",                "mandatory": False, "description": "Create a csv file with the results","biggerLabel": 20},
                {"name": "sort{Column Number}","mandatory": False, "description": "Column number to sort the table result","biggerLabel": 20},
                {"name": "desc",               "mandatory": False, "description": "Do a desceding sort","biggerLabel": 20},
                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20},
                {"name": "save",               "mandatory": False, "description": saveDescription,"biggerLabel": 20},
                {"name": "| {VALUE}",          "mandatory": False, "description": highLightDescription,"biggerLabel": 20},
                {"name": "| grep {VALUE}",     "mandatory": False, "description": filterDescription,"biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "listBuckets " + Style.RESET},
                {"command":Style.GREEN + "listBuckets " + Style.BLUE + " sort3,desc" + Style.RESET}
            ]
        },
        LIST_OBJECTS_BUCKETS_S3:{
            "group": S3,
            "name": "List objects in a bucket",
            "description": "List objects in a bucket",
            "arguments": [
                {"name": "{BUCKET}",           "mandatory": True, "description": "Name of the bucket","biggerLabel": 20},
                {"name": "{BUCKET}/{PREFIX}",  "mandatory": False, "description": "Specify a prefix, works like a \"folder\" inside S3 Buckets. If finish with a / will list it" ,"biggerLabel": 20},
                {"name": "csv",                "mandatory": False, "description": "Create a csv file with the results","biggerLabel": 20},

                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20},
                {"name": "save",               "mandatory": False, "description": saveDescription,"biggerLabel": 20},
                {"name": "sort{Column Number}","mandatory": False, "description": "Column number to sort the table result","biggerLabel": 20},
                {"name": "desc",               "mandatory": False, "description": "Do a desceding sort","biggerLabel": 20},
                {"name": "tags",               "mandatory": False, "description": "Tags to filter the resource (will be added to the Environment Tags set)","biggerLabel": 20},
                {"name": "showtags",           "mandatory": False, "description": "Add a column with all Tags of the VPC.","biggerLabel": 20},
                {"name": "| {VALUE}",          "mandatory": False, "description": highLightDescription,"biggerLabel": 20},
                {"name": "| grep {VALUE}",     "mandatory": False, "description": filterDescription,"biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "listObjectsBucketS3 " + Style.BLUE + " bucket_root" + Style.RESET},
                {"command":Style.GREEN + "listObjectsBucketS3 " + Style.BLUE + " bucket_root/folder1/" + Style.RESET},
                {"command":Style.GREEN + "listObjectsBucketS3 " + Style.BLUE + " bucket_root/folder1/folder2/" + Style.RESET},
                {"command":Style.GREEN + "listObjectsBucketS3 " + Style.BLUE + " production/logs/2019" + Style.RESET}
            ]
        },
        AWS_NAVIGATOR:{
            "group": GENERAL,
            "name": "Navigate interactively through Network information of the current AWS Account",
            "description": "Using a browser as user interface, allows to interate with Network information of the current AWS Account",
            "arguments": [
                {"name": "verbose",            "mandatory": False, "description": verboseDescription,"biggerLabel": 20}
            ],
            "examples-interactive": [
                {"command":Style.GREEN + "navigator " + Style.RESET},
                {"command":Style.GREEN + "navigator " + Style.BLUE + " verbose" + Style.RESET}
            ]
        }
    }

    def formatDescr(text, margin, width):
        MARGIN    = margin
        WIDTH     = width
        formatted = text
        line      = ""

        if len(formatted) > WIDTH:
           formatted = ""
           for word in text.split(" "):
               if len(line + word + " ") <= (WIDTH):
                  line += word + " "
               else:
                  finalLine  = Functions.fillLineSpacesUntil(line,WIDTH)
                  #finalLine  = line
                  formatted += finalLine + "\n" + "".ljust(MARGIN," ") 
                  line = word + " " 
        
        return formatted + line

    def fillLineSpacesUntil(line, width):
        finalLine = line
        posSpace  = 0
        #while True:
        #   if (len(finalLine) + 1)  < width:
        #      spaceIndex = finalLine.find(' ',posSpace)
        #      finalLine  = finalLine[:spaceIndex] + " " + finalLine[spaceIndex:]
        #      posSpace   = spaceIndex + 2
        #   else:
        #      break   
        while (len(finalLine) + 1)  < width:
            spaceIndex = finalLine.find(' ',posSpace)
            finalLine  = finalLine[:spaceIndex] + " " + finalLine[spaceIndex:]
            posSpace   = spaceIndex + 2
           
        return finalLine      

    def showFunctions():
        print (" ")
        LABEL_LENGTH_FUNCTION_NAME = 17
        SIZE_SEPARATOR             = 95
        output  = "-".ljust(SIZE_SEPARATOR,"-") + "\n"           
        output += "   🔧   " + Style.CYAN + "A W S    F U N C T I O N S" + Style.RESET + "" + "\n"
        output += "-".ljust(SIZE_SEPARATOR,"-") + "\n"    
        for key in Functions.FUNCTIONS:
            function     = Functions.FUNCTIONS[key]
            functionName = key
            functionDescr = "  👉 " + Style.GREEN + functionName + Style.RESET + (" ".ljust(LABEL_LENGTH_FUNCTION_NAME - len(functionName) + 2,"-")) + \
                                  "> "+ Functions.formatDescr(function["description"], 26, 59) + "\n"
            output += functionDescr
            output += Style.MAGENTA + "          Arguments: " + Style.RESET + "\n"
            for args in function["arguments"]:
                output += Style.BLUE + "               " + args["name"] + " ".ljust(args["biggerLabel"] - len(args["name"])," ") + Style.MAGENTA 
                output += ("[Required]" if args["mandatory"] == True else "[Optional]")
                output += Style.RESET + "....: " + Style.RESET + Functions.formatDescr(args["description"],51,45) + "\n"

            output += "          " + Style.MAGENTA + "Examples:" + Style.RESET + "\n"
            for exs in function["examples-interactive"]:
                output += "               " + exs["command"] + "\n"
            output += "-".ljust(SIZE_SEPARATOR,"-") + "\n"
        print(output)

    def showFunctionsSummary():
        print (" ")
        LABEL_LENGTH_FUNCTION_NAME = 23
        SIZE_SEPARATOR             = 95
        output  = "-".ljust(SIZE_SEPARATOR,"-") + "\n"           
        output += "   🔧   " + Style.CYAN + "A W S    F U N C T I O N S" + Style.RESET + "" + "\n"
        output += "-".ljust(SIZE_SEPARATOR,"-") + "\n"    
        for key in Functions.FUNCTIONS:
            function     = Functions.FUNCTIONS[key]
            functionName = key
            functionDescr = "  👉 " + Style.GREEN + functionName + Style.RESET + (" ".ljust(LABEL_LENGTH_FUNCTION_NAME - len(functionName) + 2,"-")) + \
                                  "> "+ Functions.formatDescr(function["description"], 32, 59) + "\n"
            output += functionDescr
        print(output)         


Functions.formatDescr         = staticmethod(Functions.formatDescr)
Functions.fillLineSpacesUntil = staticmethod(Functions.fillLineSpacesUntil)
Functions.showFunctions       = staticmethod(Functions.showFunctions)


import boto3

class Config:

    def __init__(self):
        self.awsProfile               = ""
        self.awsRegion                = ""
        self.awsTags                  = {}
        self.command                  = ""
        self.commandArguments         = ""
        self.tableLineSeparator       = True
        self.interactive              = False
        self.printResults             = False
        self.uploadChunkSizeMultipart = 10
        self.uploadThresholdMultipart = 100

    def awsTagsToFilter(self):
        if self.awsTags and len(self.awsTags) > 0:
            return True
        return False

    def botoSession(self):
        if not self.awsProfile and not self.awsRegion:
           session = boto3.Session()
        elif not self.awsProfile and     self.awsRegion:    
           session = boto3.Session(region_name=self.awsRegion)
        elif     self.awsProfile and not self.awsRegion:   
           session = boto3.Session(profile_name=self.awsProfile)
        else:
           session = boto3.Session(profile_name=self.awsProfile,region_name=self.awsRegion) 
        return session        

    




# coding=utf-8
#
# Author: Ualter Otoni Pereira
# ualter.junior@gmail.com
#

import os, sys
import math
from pyawscp.Utils import Utils, Style

class Leaf:
    def __init__(self, label):
        self.order       = 0
        self.label       = label
        self.leaves      = []
        self.parent      = None
        self.colSize     = 0
        self.marginSize  = 0
        self.level       = 0
        self.center      = 0
        self.widthBox    = 0
        self.offset      = 0
        self.prevSibling = None
        self.nextSibling = None

    def add(self, leaf):
        leaf.parent = self
        if len(self.leaves) > 0:
           self.leaves[-1].nextSibling = leaf
           leaf.prevSibling = self.leaves[-1]
        self.leaves.append(leaf) 
    def childLeaves(self):
        return self.leaves
    def qtdeChildLeaves(self):
        return len(self.leaves)
    def hasChildren(self):
        if self.qtdeChildLeaves() > 0:
           return True
        return False
    def myRealSize(self):
        return int(self.marginSize + self.widthBox + self.offset)


class Layout:
    def __init__(self, layout):
       self._layout           = layout
       self._background       = [" "," "]
       #self._background       = ["•","░"]
       self._fill             = [" "," "]
       self._lateralBar       = ["|","│"]
       self._line             = ["-","─"]  
       self._topRightCorner   = ["+","┐"]
       self._topLeftCorner    = ["+","┌"]
       self._belowRightCorner = ["+","┘"]
       self._belowLeftCorner  = ["+","└"]
       
    def fill(self):
       return self._fill[self._layout-1]
    def lateralBar(self):
       return str(self._lateralBar[self._layout-1])
    def line(self):
       return str(self._line[self._layout-1])      
    def topRightCorner(self):
       return self._topRightCorner[self._layout-1]
    def topLeftCorner(self):
       return self._topLeftCorner[self._layout-1]   
    def belowRightCorner(self):
       return self._belowRightCorner[self._layout-1]
    def belowLeftCorner(self):
       return self._belowLeftCorner[self._layout-1]      
    def background(self):
       return self._background[self._layout-1]   
        
class PyAsciiGraphs:
    def __init__(self):
        self.layoutOpt  = 2
        self.structure  = ""
        _rows, _columns = os.popen('stty size', 'r').read().split()
        self.columns    = int(_columns)
        self.rows       = int(_rows)
        self.center     = int(self.columns / 2)
        self.layout     = Layout(self.layoutOpt)

        self.tree            = {} 
        self.treeSplitColumn = 0
        self.treeQtdeColumn  = 0

    def replicate(self,qtde,chars=" "):
        return chars.ljust(qtde,chars)

    def drawBox(self, labelContent, _cols=None, offset=0, printMe=True):
        lines = []
        if not _cols:
           cols = self.columns
        else:
           cols = _cols   
        center  = int(cols / 2)

        widthBox    = len(labelContent) + 2
        margin      = "".ljust(cols - (int(widthBox / 2) + center),self.layout.background())
        margin     += "".ljust(offset,self.layout.background())
        topLine     = margin + self.layout.topLeftCorner() + ( self.replicate(widthBox,self.layout.line())  ) + self.layout.topRightCorner()
        middleLine  = margin + self.layout.lateralBar() + self.layout.fill() + Style.GREEN + labelContent + Style.RESET + self.layout.fill() + self.layout.lateralBar()
        bottomLine  = margin + self.layout.belowLeftCorner() + ( self.replicate(widthBox,self.layout.line()) ) + self.layout.belowRightCorner()

        lines.append(topLine)
        lines.append(middleLine)
        lines.append(bottomLine)
        if printMe:
           for l in lines:
               print(l)

    def createTree(self,rootLeaf):
        self.createTreeLeaves(rootLeaf,0)

    def labelSize(self,label,length):
        return label + "".ljust(length-len(label)," ")

    def printTreeData(self):
        header  = self.labelSize("Label",30) + " | " + self.labelSize("Parent",25) + " | " + \
                 self.labelSize("Level",10) + " | " + self.labelSize("Order",10) + " | " + self.labelSize("ColSize",10) + " | " + \
                 self.labelSize("MarginSize",10) + " | " + self.labelSize("Center",10) + " | " + self.labelSize("WidthBox",10)
        print(header)
        for level in self.tree:
            for order in self.tree[level]:
                label      = self.labelSize(level + "," + order + "--> " + self.tree[level][order]["leaf"].label,30)
                parent     = self.labelSize(self.tree[level][order]["leaf"].parent.label if self.tree[level][order]["leaf"].parent else " ",25)
                _level     = self.labelSize(str(self.tree[level][order]["leaf"].level),10)
                _order     = self.labelSize(str(self.tree[level][order]["leaf"].order),10)
                colSize    = self.labelSize(str(self.tree[level][order]["leaf"].colSize),10)
                center     = self.labelSize(str(self.tree[level][order]["leaf"].center),10)
                marginSize = self.labelSize(str(self.tree[level][order]["leaf"].marginSize),10)
                widthBox   = self.labelSize(str(len(self.tree[level][order]["leaf"].label) +  2),10)

                line  = label + " | " + parent + " | " + _level + " | " + _order + " | " + colSize + " | " + marginSize + " | " + center + " | " + widthBox
                print(line)
            print("-".ljust(150,"-"))    

    def printIds(self,parent)         :
        for leaf in parent.leaves:
            label = leaf.label + "".ljust(8-len(leaf.label)," ")
            print(label + " level=" + str(leaf.level) + ", order=" + str(leaf.order))   
            self.printIds(leaf)

    def createTreeLeaves(self,parent,level):
        if str(level) not in self.tree:
           self.tree[str(level)] = {}
        parent.level = level
        #print(parent.label + ", level=" + str(parent.level) + ", order=" + str(parent.order))
        level += 1
        order  = 1
        if len(parent.leaves):
           if self.treeSplitColumn == 0 and len(parent.leaves) > 1:
              self.treeQtdeColumn  = len(parent.leaves)
              self.treeSplitColumn = level
           for leaf in parent.leaves:
               leaf.level = level
               leaf.order = order
               
               if str(level) not in self.tree:
                  self.tree[str(level)] = {}
               self.tree[str(level)][str(parent.order) + "." + str(order)] = {"leaf":leaf}

               #print(" ------> " + leaf.label + ", level=" + str(leaf.level) + ", order=" + str(leaf.order))
               self.createTreeLeaves(leaf,level)
               order += 1

    def calculateOffsetLeaf(self, leaf):
        # First get the level of the first column split of the tree, where all started to be in separated columns   
        orderParent     = 0
        orderGranParent = 0
        l               = leaf
        while True:
            if l.parent and l.parent.level == self.treeSplitColumn:
               orderParent = l.parent.order
               if l.parent.parent: 
                  orderGranParent = l.parent.parent.order
               break
            if not l.parent:
               break
            l = l.parent

        # Once found, if found, check if the neighbor main column has something printed or is empty
        if orderParent > 0:
           orderParentNeighbor = int(orderParent) - 1

           # Check if there is a neighbor for its parent
           keyLeaf = str(orderGranParent) + "." + str(orderParentNeighbor)
           if keyLeaf in self.tree[str(self.treeSplitColumn)]:
              parentNeighborLeaf = self.tree[str(self.treeSplitColumn)][str(keyLeaf)]["leaf"]
              colSizeNeighbor    = parentNeighborLeaf.colSize
              hasChildren        = parentNeighborLeaf.hasChildren()
              if not hasChildren:
                 offset = colSizeNeighbor * orderParentNeighbor
                 #print("leaf:" + leaf.label +  ", orderParent=" + str(orderParent) +  ", colSizeNeighbor=" + str(colSizeNeighbor) +  ", hasChildren=" + str(hasChildren)\
                 #    +  ", offset=" + str(offset)) 
                 return offset
              else:
                 # Sum all the spaces occupied - the column total size  (Starting from the )
                 if (leaf.level - parentNeighborLeaf.level) > 1: 
                    # Not Ok, we should go down until find the rigth above level (diff 1, between Level child and parent)
                    # Go down 1 level below
                    levelsBelow = 1
                    nextLevel   = parentNeighborLeaf.level + 1
                    while  (leaf.level - nextLevel) > 1:
                      nextLevel   += 1
                      levelsBelow += 1
                    keyLeaf = str(orderGranParent + levelsBelow) + "." + str(orderParentNeighbor)
                    while keyLeaf not in self.tree[str(nextLevel)]:
                        # In case there's no neighbor direct parent, we go again in reverse direction (going up) to find the first neighbor parent level with leaves
                        nextLevel -= 1
                    parentNeighborLeaf = self.tree[str(nextLevel)][str(keyLeaf)]["leaf"]

                 # Calculate how many offset according the sum of colsizes of the children neighbor, 
                 # minus the colsize of the parent neighbor (difference)
                 spaceOcuppiedChildren = 0
                 for l in parentNeighborLeaf.leaves:
                     #spaceOcuppiedChildren += l.colSize
                     spaceOcuppiedChildren += l.myRealSize()
                 offset = parentNeighborLeaf.colSize - spaceOcuppiedChildren

                 # Just in case
                 if offset < 0:
                    offset = 0
                 return offset
        return 0    

    def offSetByCalculatingSpaceNotOccupiedByChildren(self, parentLeaf):
        spaceOcuppiedChildren = 0
        for l in parentLeaf.leaves:
            spaceOcuppiedChildren += l.myRealSize()
        offset = parentLeaf.colSize - spaceOcuppiedChildren
        return offset

    def calculateLayout(self):
        for _level in self.tree:
           if int(_level) > 0:
              for _order in self.tree[str(_level)]:
                  leaf        = self.tree[str(_level)][str(_order)]["leaf"]
                  #qtdeColumns = len(self.tree[str(_level)])
                  qtdeColumns = len(leaf.parent.leaves)

                  if int(_level) == 1:
                     colSpace = math.ceil( self.columns / qtdeColumns )
                  else:
                     parent   = leaf.parent
                     colSpace = math.ceil( parent.colSize /  qtdeColumns )   

                  center          = math.ceil(colSpace / 2)
                  widthBox        = len(Utils.removeCharsColors(leaf.label)) + 2
                  marginSize      = colSpace - (math.ceil(widthBox / 2) + center)
                  leaf.center     = center
                  leaf.colSize    = colSpace
                  leaf.widthBox   = widthBox
                  leaf.marginSize = marginSize

    def drawTree(self,rootLeaf):
        self.createTree(rootLeaf)
        self.calculateLayout()
        #self.printTreeData()
        # Only the root, the simplest thing
        self.drawBox(rootLeaf.label)
        # Print the Tree
        Matrix = {}
        for level in self.tree:
            totalColumnSize = 0
            Matrix[str(level)] = {}
            Matrix[str(level)]["topLine"]    = ""
            Matrix[str(level)]["middleLine"] = ""
            Matrix[str(level)]["bottomLine"] = ""
            for _order in self.tree[level]:
                leaf        = self.tree[str(level)][str(_order)]["leaf"]
                parentOrder = int(_order.split(".")[0])
                order       = int(_order.split(".")[1])

                #print(leaf.label)
                offset = 0
                if int(level) == self.treeSplitColumn:
                   # Second Line after root
                   baseColSize = leaf.colSize
                   
                if int(order) > 1 and int(level) == self.treeSplitColumn:
                   # how many main columns is built the tree (the first line that split in columns)
                   remainingSize = (baseColSize * (int(order)-1)) - totalColumnSize
                   if remainingSize > 0:
                      offset = offset + remainingSize  
                else:
                   # For the rest of levels, we need to check the "neighbor" column spaces printed (based on the level 1)
                   # Because if nothing is there, we need to "paint" the empty space to let the position properly idented
                   if int(order) == 1:
                      offset = self.calculateOffsetLeaf(leaf)
                   else:
                      diff = leaf.colSize - (len(leaf.label) + 2 + leaf.marginSize)   
                      offset = self.calculateOffsetLeaf(leaf)
                      if diff > 0:
                         offset = diff

                #print(leaf.label)
                #print("order="+str(order)) 
                #print("marginSize="+str(offset)) 
                #print("offset="+str(offset)) 
                #print("baseColSize="+str(baseColSize)) 
                #print("totalColumnSize="+str(totalColumnSize)) 
                #print("---")

                #marginSpaces = "|".ljust(leaf.marginSize - 1,self.layout.background())
                marginSpaces = "".ljust(leaf.marginSize,self.layout.background())
                offsetSpaces = "".ljust(offset,self.layout.background())

                leaf.offset = offset

                topLine     = offsetSpaces + marginSpaces + self.layout.topLeftCorner() + ( self.replicate(leaf.widthBox,self.layout.line())  ) + self.layout.topRightCorner()
                middleLine  = offsetSpaces + marginSpaces + self.layout.lateralBar() + self.layout.fill() + Style.GREEN + leaf.label + Style.RESET + self.layout.fill() + self.layout.lateralBar()
                bottomLine  = offsetSpaces + marginSpaces + self.layout.belowLeftCorner() + ( self.replicate(leaf.widthBox,self.layout.line()) ) + self.layout.belowRightCorner()

                Matrix[str(level)]["topLine"]    += topLine
                Matrix[str(level)]["middleLine"] += middleLine
                Matrix[str(level)]["bottomLine"] += bottomLine

                totalColumnSize += len(Utils.removeCharsColors(middleLine))
                #print(middleLine + "| --> " + str(totalColumnSize))

        for l in Matrix:
            if len(Matrix[l]["topLine"]) > 0:
               print(Matrix[l]["topLine"])
               print(Matrix[l]["middleLine"])
               print(Matrix[l]["bottomLine"])
        # End Print the Tree

def testTree1():
    pygraph = PyAsciiGraphs()

    rootLeaf = Leaf("Root")

    l1_0 = Leaf("1.0")
    l2_0 = Leaf("2.0")
    l3_0 = Leaf("3.0")

    l1_1 = Leaf("1.1")
    l1_2 = Leaf("1.2")

    l2_1 = Leaf("2.1")
    l2_2 = Leaf("2.2")
    l2_3 = Leaf("2.3")

    l3_1 = Leaf("3.1")
    l3_2 = Leaf("3.2")

    l1_0.add(l1_1)
    l1_0.add(l1_2)

    l2_0.add(l2_1)
    l2_0.add(l2_2)
    l2_0.add(l2_3)

    l3_0.add(l3_1)
    l3_0.add(l3_2)

    rootLeaf.add(l1_0)
    rootLeaf.add(l2_0)
    rootLeaf.add(l3_0)

    pygraph.drawTree(rootLeaf)

def test2():
    pygraph = PyAsciiGraphs()

    rootLeaf = Leaf("services.com.uk")
    site1 = Leaf("site1.services.com.uk")
    site2 = Leaf("site2.services.com.uk")
    site1London = Leaf("london.site2.services.com.uk")
    site2irish  = Leaf("irih.site2.services.com.uk")
    site2wales  = Leaf("wals.site2.services.com.uk")


    #site1.add(site1London)
    site1.add(Leaf("liverpool.site2.services.com.uk"))

    site2dublin = Leaf("dublin.irih.site2.services.com.uk")   
    site2irish.add(site2dublin) 

    site2.add(site2irish)
    site2.add(site2wales)
    rootLeaf.add(site1)
    rootLeaf.add(site2)
    pygraph.drawTree(rootLeaf)

def test3():
    pygraph = PyAsciiGraphs()

    rootLeaf = Leaf("AAAAAA")
    site1 = Leaf("BBBBB")
    site2 = Leaf("CCCCC")
    site3 = Leaf("DDDDD")

    b10 = Leaf("B1.0")
    b20 = Leaf("B2.0")
    #site1.add(b10)
    #site1.add(b20)

    d10 = Leaf("D1.0")
    d101 = Leaf("D1.0.1")
    d102 = Leaf("D1.0.2")
    dx   = Leaf("D1.0.1.1")
    d101.add(dx)
    d10.add(d101)
    d10.add(d102)
    d20 = Leaf("D2.0")
    site3.add(d10)
    site3.add(d20)

    rootLeaf.add(site1)
    rootLeaf.add(site2)
    rootLeaf.add(site3)
    pygraph.drawTree(rootLeaf)

def test4():
    pygraph = PyAsciiGraphs()

    rootLeaf = Leaf("services.com.uk")
    site1 = Leaf("site1.services.com.uk")

    elb = Leaf("elb")

    target1 = Leaf("Target Group 1")
    target2 = Leaf("Target Group 2")

    ec1 = Leaf("EC2 1")
    ec2 = Leaf("EC2 2")
    ec3 = Leaf("EC2 3")
    ec4 = Leaf("EC2 4")

    target1.add(ec1)
    target1.add(ec2)
    target2.add(ec3)
    target2.add(ec4)
    elb.add(target1)
    elb.add(target2)

    site1.add(elb)
    rootLeaf.add(site1)
    #if pygraph.columns > 150:
    #   pygraph.columns = 150
    pygraph.drawTree(rootLeaf)

def test5():
    pygraph = PyAsciiGraphs()

    rootLeaf = Leaf("Root")
    l1_0 = Leaf("1.0")
    l2_0 = Leaf("2.0")

    l1_1 = Leaf("1.1")
    l1_0.add(l1_1)

    l2_1 = Leaf("2.1")
    l2_2 = Leaf("2.2")
    l2_3 = Leaf("2.3")
    l2_4 = Leaf("2.4")
    l2_1_1 = Leaf("2.1.1")
    l2_1_2 = Leaf("2.1.2")
    
    l2_1_2.add(Leaf("2.1.2.1"))

    l2_1.add(l2_1_1)
    l2_1.add(l2_1_2)

    l2_0.add(l2_1)
    l2_0.add(l2_2)
    l2_0.add(l2_3)
    l2_0.add(l2_4)

    rootLeaf.add(l1_0)
    rootLeaf.add(l2_0)
    rootLeaf.add(Leaf("3.0"))
    pygraph.drawTree(rootLeaf)

def test6():
    pygraph = PyAsciiGraphs()

    rootLeaf = Leaf("Root")
    l1_0 = Leaf("1.0")
    l2_0 = Leaf("2.0")
    l3_0 = Leaf("3.0")

    #l1_0.add(Leaf("1.1"))

    l3_1 = Leaf("3.1")
    l3_2 = Leaf("3.2")
    l3_3 = Leaf("3.3")
    l3_4 = Leaf("3.4")
    l3_1_1 = Leaf("3.1.1")
    l3_1_2 = Leaf("3.1.2")
    
    l3_1_2.add(Leaf("3.1.2.1"))

    l3_1.add(l3_1_1)
    l3_1.add(l3_1_2)

    l3_0.add(l3_1)
    l3_0.add(l3_2)
    l3_0.add(l3_3)
    l3_0.add(l3_4)

    rootLeaf.add(l1_0)
    rootLeaf.add(l2_0)
    rootLeaf.add(l3_0)
    pygraph.columns = 188
    pygraph.drawTree(rootLeaf)    

if __name__ == '__main__':
    testTree1()
    #print("\n**********************************************************************************************\n")
    #test2()
    #print("\n**********************************************************************************************\n")
    #test4()
    #print("\n**********************************************************************************************\n")
    #test5()
    #print("\n**********************************************************************************************\n")
    #test6()
   

    
    

        






